# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Sets the environment used by the python-oracledb test suite. Production
# applications should consider using External Authentication to avoid hard
# coded credentials.
#
# You can set values in environment variables to bypass having the test suite
# request the information it requires.
#
#   PYO_TEST_MAIN_USER: user used for most test cases
#   PYO_TEST_MAIN_PASSWORD: password of user used for most test cases
#   PYO_TEST_PROXY_USER: user for testing proxying
#   PYO_TEST_PROXY_PASSWORD: password of user for testing proxying
#   PYO_TEST_CONNECT_STRING: connect string for test suite
#   PYO_TEST_ADMIN_USER: administrative user for test suite
#   PYO_TEST_ADMIN_PASSWORD: administrative password for test suite
#   PYO_TEST_WALLET_LOCATION: location of wallet file (thin mode, mTLS)
#   PYO_TEST_WALLET_PASSWORD: password for wallet file (thin mode, mTLS)
#   PYO_TEST_DRIVER_MODE: python-oracledb mode (thick or thin) to use
#   PYO_TEST_EXTERNAL_USER: user for testing external authentication
#
# PYO_TEST_CONNECT_STRING can be set to an Easy Connect string, or a
# Net Service Name from a tnsnames.ora file or external naming service,
# or it can be the name of a local Oracle database instance.
#
# If oracledb is using Instant Client, then an Easy Connect string is generally
# appropriate. The syntax is:
#
#   [//]host_name[:port][/service_name][:server_type][/instance_name]
#
# Commonly just the host_name and service_name are needed
# e.g. "localhost/orclpdb1" or "localhost/XEPDB1"
#
# If using a tnsnames.ora file, the file can be in a default
# location such as $ORACLE_HOME/network/admin/tnsnames.ora or
# /etc/tnsnames.ora.  Alternatively set the TNS_ADMIN environment
# variable and put the file in $TNS_ADMIN/tnsnames.ora.
#
# The administrative user for cloud databases is ADMIN and the administrative
# user for on premises databases is SYSTEM.
# -----------------------------------------------------------------------------

import getpass
import os
import sys
import unittest

import oracledb

# Python 3.7 doesn't have support for testing asyncio so fake it to avoid the
# entire test suite failing!
if not hasattr(unittest, "IsolatedAsyncioTestCase"):
    unittest.IsolatedAsyncioTestCase = object

# default values
DEFAULT_MAIN_USER = "pythontest"
DEFAULT_PROXY_USER = "pythontestproxy"
DEFAULT_CONNECT_STRING = "localhost/orclpdb1"

# dictionary containing all parameters; these are acquired as needed by the
# methods below (which should be used instead of consulting this dictionary
# directly) and then stored so that a value is not requested more than once
PARAMETERS = {}


def get_value(name, label, default_value=None, password=False):
    try:
        return PARAMETERS[name]
    except KeyError:
        pass
    env_name = "PYO_TEST_" + name
    value = os.environ.get(env_name)
    if value is None:
        if default_value is not None:
            label += " [%s]" % default_value
        label += ": "
        if password:
            value = getpass.getpass(label)
        else:
            value = input(label).strip()
    if not value:
        value = default_value
    PARAMETERS[name] = value
    return value


def get_admin_connection(use_async=False):
    if not get_is_thin() and oracledb.is_thin_mode():
        oracledb.init_oracle_client()
    admin_user = get_value("ADMIN_USER", "Administrative user", "admin")
    admin_password = get_value(
        "ADMIN_PASSWORD", f"Password for {admin_user}", password=True
    )
    params = get_connect_params()
    if admin_user and admin_user.upper() == "SYS":
        params = params.copy()
        params.set(mode=oracledb.AUTH_MODE_SYSDBA)
    method = oracledb.connect_async if use_async else oracledb.connect
    return method(
        dsn=get_connect_string(),
        params=params,
        user=admin_user,
        password=admin_password,
    )


async def get_admin_connection_async(use_async=False):
    return await get_admin_connection(use_async=True)


def get_charset():
    value = PARAMETERS.get("CHARSET")
    if value is None:
        with get_connection() as conn:
            with conn.cursor() as cursor:
                cursor.execute(
                    """
                    select value
                    from nls_database_parameters
                    where parameter = 'NLS_CHARACTERSET'
                    """
                )
                (value,) = cursor.fetchone()
                PARAMETERS["CHARSET"] = value
    return value


async def get_charset_async():
    value = PARAMETERS.get("CHARSET")
    if value is None:
        async with get_connection_async() as conn:
            with conn.cursor() as cursor:
                await cursor.execute(
                    """
                    select value
                    from nls_database_parameters
                    where parameter = 'NLS_CHARACTERSET'
                    """
                )
                (value,) = await cursor.fetchone()
                PARAMETERS["CHARSET"] = value
    return value


def get_charset_ratios():
    value = PARAMETERS.get("CS_RATIO")
    if value is None:
        connection = get_connection()
        cursor = connection.cursor()
        cursor.execute(
            """
            select
                cast('X' as varchar2(1)),
                cast('Y' as nvarchar2(1))
            from dual
            """
        )
        varchar_column_info, nvarchar_column_info = cursor.description
        value = (varchar_column_info[3], nvarchar_column_info[3])
        PARAMETERS["CS_RATIO"] = value
    return value


async def get_charset_ratios_async():
    value = PARAMETERS.get("CS_RATIO")
    if value is None:
        connection = await get_connection_async()
        cursor = connection.cursor()
        await cursor.execute(
            """
            select
                cast('X' as varchar2(1)),
                cast('Y' as nvarchar2(1))
            from dual
            """
        )
        varchar_column_info, nvarchar_column_info = cursor.description
        value = (varchar_column_info[3], nvarchar_column_info[3])
        PARAMETERS["CS_RATIO"] = value
    return value


def get_client_version():
    name = "CLIENT_VERSION"
    value = PARAMETERS.get(name)
    if value is None:
        if get_is_thin():
            value = (23, 5)
        else:
            oracledb.init_oracle_client()
            value = oracledb.clientversion()[:2]
        PARAMETERS[name] = value
    return value


def get_connect_params():
    wallet_location = get_wallet_location()
    return oracledb.ConnectParams(
        user=get_main_user(),
        password=get_main_password(),
        config_dir=wallet_location,
        wallet_location=wallet_location,
        wallet_password=get_wallet_password(),
        disable_oob=True,
    )


def get_connection(dsn=None, use_async=False, **kwargs):
    if not get_is_thin() and oracledb.is_thin_mode():
        oracledb.init_oracle_client()
    if dsn is None:
        dsn = get_connect_string()
    method = oracledb.connect_async if use_async else oracledb.connect
    return method(dsn=dsn, params=get_connect_params(), **kwargs)


def get_connection_async(dsn=None, **kwargs):
    return get_connection(dsn, use_async=True, **kwargs)


def get_connect_string():
    return get_value(
        "CONNECT_STRING", "Connect String", DEFAULT_CONNECT_STRING
    )


def get_is_drcp():
    value = PARAMETERS.get("IS_DRCP")
    if value is None:
        params = oracledb.ConnectParams()
        params.parse_connect_string(get_connect_string())
        server_type = params.server_type
        value = (
            server_type == "pooled"
            or isinstance(server_type, list)
            and "pooled" in server_type
        )
        PARAMETERS["IS_DRCP"] = value
    return value


def get_is_implicit_pooling():
    value = PARAMETERS.get("IS_IMPLICIT_POOLING")
    if value is None:
        if not get_is_drcp():
            value = False
        else:
            params = oracledb.ConnectParams()
            params.parse_connect_string(get_connect_string())
            pool_boundary = params.pool_boundary
            value = (
                pool_boundary is not None
                or isinstance(pool_boundary, list)
                and [s for s in pool_boundary if s]
            )
        PARAMETERS["IS_IMPLICIT_POOLING"] = value
    return value


def get_is_thin():
    driver_mode = get_value("DRIVER_MODE", "Driver mode (thin|thick)", "thin")
    return driver_mode == "thin"


def get_main_password():
    return get_value(
        "MAIN_PASSWORD", f"Password for {get_main_user()}", password=True
    )


def get_main_user():
    return get_value("MAIN_USER", "Main User Name", DEFAULT_MAIN_USER)


def get_pool(use_async=False, **kwargs):
    method = oracledb.create_pool_async if use_async else oracledb.create_pool
    return method(dsn=get_connect_string(), params=get_pool_params(), **kwargs)


def get_pool_async(**kwargs):
    return get_pool(use_async=True, **kwargs)


def get_pool_params():
    wallet_location = get_wallet_location()
    return oracledb.PoolParams(
        user=get_main_user(),
        password=get_main_password(),
        config_dir=wallet_location,
        wallet_location=wallet_location,
        wallet_password=get_wallet_password(),
    )


def get_proxy_password():
    return get_value(
        "PROXY_PASSWORD", f"Password for {get_proxy_user()}", password=True
    )


def get_proxy_user():
    return get_value("PROXY_USER", "Proxy User Name", DEFAULT_PROXY_USER)


def get_sleep_proc_name():
    server_version = get_server_version()
    return (
        "dbms_session.sleep" if server_version[0] >= 18 else "dbms_lock.sleep"
    )


def get_server_version():
    name = "SERVER_VERSION"
    value = PARAMETERS.get(name)
    if value is None:
        conn = get_connection()
        value = tuple(int(s) for s in conn.version.split("."))[:2]
        PARAMETERS[name] = value
    return value


async def get_server_version_async():
    name = "SERVER_VERSION"
    value = PARAMETERS.get(name)
    if value is None:
        async with await get_connection_async() as conn:
            value = tuple(int(s) for s in conn.version.split("."))[:2]
            PARAMETERS[name] = value
    return value


def get_wallet_location():
    if get_is_thin():
        return get_value("WALLET_LOCATION", "Wallet Location")


def get_wallet_password():
    if get_is_thin():
        return get_value("WALLET_PASSWORD", "Wallet Password", password=True)


def get_external_user():
    if not get_is_thin():
        return get_value("EXTERNAL_USER", "External User")


def is_on_oracle_cloud(connection):
    server = get_server_version()
    if server < (18, 0):
        return False
    cursor = connection.cursor()
    cursor.execute(
        """
        select sys_context('userenv', 'cloud_service')
        from dual
        """
    )
    (service_name,) = cursor.fetchone()
    return service_name is not None


async def is_on_oracle_cloud_async(connection):
    server = await get_server_version_async()
    if server < (18, 0):
        return False
    cursor = connection.cursor()
    await cursor.execute(
        """
        select sys_context('userenv', 'cloud_service')
        from dual
        """
    )
    (service_name,) = await cursor.fetchone()
    return service_name is not None


def run_sql_script(conn, script_name, **kwargs):
    statement_parts = []
    cursor = conn.cursor()
    replace_values = [("&" + k + ".", v) for k, v in kwargs.items()] + [
        ("&" + k, v) for k, v in kwargs.items()
    ]
    script_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
    file_name = os.path.join(script_dir, "sql", script_name + ".sql")
    for line in open(file_name):
        if line.strip() == "/":
            statement = "".join(statement_parts).strip()
            if statement:
                for search_value, replace_value in replace_values:
                    statement = statement.replace(search_value, replace_value)
                try:
                    cursor.execute(statement)
                except:
                    print("Failed to execute SQL:", statement)
                    raise
            statement_parts = []
        else:
            statement_parts.append(line)
    cursor.execute(
        """
        select name, type, line, position, text
        from dba_errors
        where owner = upper(:owner)
        order by name, type, line, position
        """,
        owner=get_main_user(),
    )
    prev_name = prev_obj_type = None
    for name, obj_type, line_num, position, text in cursor:
        if name != prev_name or obj_type != prev_obj_type:
            print("%s (%s)" % (name, obj_type))
            prev_name = name
            prev_obj_type = obj_type
        print("    %s/%s %s" % (line_num, position, text))


def run_test_cases():
    unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))


def skip_soda_tests():
    if get_is_thin():
        return True
    client = get_client_version()
    if client < (18, 3):
        return True
    server = get_server_version()
    if server < (18, 0):
        return True
    if server > (20, 1) and client < (20, 1):
        return True
    return False


class DefaultsContextManager:
    def __init__(self, attribute, desired_value):
        self.attribute = attribute
        self.desired_value = desired_value

    def __enter__(self):
        self.original_value = getattr(oracledb.defaults, self.attribute)
        setattr(oracledb.defaults, self.attribute, self.desired_value)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        setattr(oracledb.defaults, self.attribute, self.original_value)


class SystemStatInfo:
    get_sid_sql = "select sys_context('userenv', 'sid') from dual"
    get_stat_sql = """
          select ss.value
          from v$sesstat ss, v$statname sn
          where ss.sid = :sid
              and ss.statistic# = sn.statistic#
              and sn.name = :stat_name
          """
    stat_name = None

    def _initialize(self, connection):
        self.prev_value = 0
        self.admin_conn = get_admin_connection()
        with connection.cursor() as cursor:
            cursor.execute(self.get_sid_sql)
            (self.sid,) = cursor.fetchone()
        self.get_value()

    async def _initialize_async(self, connection):
        self.prev_value = 0
        self.admin_conn = await get_admin_connection_async()
        with connection.cursor() as cursor:
            await cursor.execute(self.get_sid_sql)
            (self.sid,) = await cursor.fetchone()
        await self.get_value_async()

    def get_value(self):
        with self.admin_conn.cursor() as cursor:
            cursor.execute(
                self.get_stat_sql, sid=self.sid, stat_name=self.stat_name
            )
            (current_value,) = cursor.fetchone()
            diff_value = current_value - self.prev_value
            self.prev_value = current_value
            return diff_value

    async def get_value_async(self):
        with self.admin_conn.cursor() as cursor:
            await cursor.execute(
                self.get_stat_sql, sid=self.sid, stat_name=self.stat_name
            )
            (current_value,) = await cursor.fetchone()
            diff_value = current_value - self.prev_value
            self.prev_value = current_value
            return diff_value


class RoundTripInfo(SystemStatInfo):
    stat_name = "SQL*Net roundtrips to/from client"


class ParseCountInfo(SystemStatInfo):
    stat_name = "parse count (total)"


class FullCodeErrorContextManager:

    def __init__(self, full_codes):
        self.full_codes = full_codes
        if len(full_codes) == 1:
            self.message_fragment = f'Error "{full_codes[0]}"'
        else:
            message_fragment = ", ".join(f'"{s}"' for s in full_codes[:-1])
            message_fragment += f' or "{full_codes[-1]}"'
            self.message_fragment = f"One of the errors {message_fragment}"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            raise AssertionError(f"{self.message_fragment} was not raised.")
        if not issubclass(exc_type, oracledb.Error):
            return False
        if issubclass(exc_type, oracledb.Error):
            error_obj = exc_value.args[0]
            if error_obj.full_code not in self.full_codes:
                message = (
                    f"{self.message_fragment} should have been raised but "
                    f'"{error_obj.full_code}" was raised instead.'
                )
                raise AssertionError(message)
        return True


class BaseTestCase(unittest.TestCase):
    requires_connection = True

    def assertParseCount(self, n):
        self.assertEqual(self.parse_count_info.get_value(), n)

    def assertRaisesFullCode(self, *full_codes):
        return FullCodeErrorContextManager(full_codes)

    def assertRoundTrips(self, n):
        self.assertEqual(self.round_trip_info.get_value(), n)

    def get_and_clear_queue(
        self,
        queue_name,
        payload_type=None,
        message="not supported with this client/server combination",
    ):
        if payload_type == "JSON":
            min_version = (21, 0)
            if (
                get_client_version() < min_version
                or get_server_version() < min_version
            ):
                self.skipTest(message)
        elif isinstance(payload_type, str):
            payload_type = self.conn.gettype(payload_type)
        queue = self.conn.queue(queue_name, payload_type)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        while queue.deqone():
            pass
        return self.conn.queue(queue_name, payload_type)

    def get_db_object_as_plain_object(self, obj):
        if obj.type.iscollection:
            element_values = []
            for value in obj.aslist():
                if isinstance(value, oracledb.DbObject):
                    value = self.get_db_object_as_plain_object(value)
                elif isinstance(value, oracledb.LOB):
                    value = value.read()
                element_values.append(value)
            return element_values
        attr_values = []
        for attribute in obj.type.attributes:
            value = getattr(obj, attribute.name)
            if isinstance(value, oracledb.DbObject):
                value = self.get_db_object_as_plain_object(value)
            elif isinstance(value, oracledb.LOB):
                value = value.read()
            attr_values.append(value)
        return tuple(attr_values)

    def get_sid_serial(self, conn=None):
        """
        Returns the sid and serial number of the connection as a 2-tuple.
        """
        if conn is None:
            conn = self.conn
        with conn.cursor() as cursor:
            cursor.execute(
                """
                select
                    dbms_debug_jdwp.current_session_id,
                    dbms_debug_jdwp.current_session_serial
                from dual
                """
            )
            return cursor.fetchone()

    def get_soda_database(
        self,
        minclient=(18, 3),
        minserver=(18, 0),
        message="not supported with this client/server combination",
        drop_collections=True,
    ):
        client = get_client_version()
        if client < minclient:
            self.skipTest(message)
        server = get_server_version()
        if server < minserver:
            self.skipTest(message)
        if server > (20, 1) and client < (20, 1):
            self.skipTest(message)
        soda_db = self.conn.getSodaDatabase()
        if drop_collections:
            for name in soda_db.getCollectionNames():
                soda_db.openCollection(name).drop()
        return soda_db

    def is_on_oracle_cloud(self, connection=None):
        if connection is None:
            connection = self.conn
        return is_on_oracle_cloud(connection)

    def setUp(self):
        if not get_is_thin() and oracledb.is_thin_mode():
            oracledb.init_oracle_client()
        if self.requires_connection:
            self.conn = get_connection()
            self.cursor = self.conn.cursor()
            self.cursor.execute("alter session set time_zone = '+00:00'")

    def setup_parse_count_checker(self, conn=None):
        if get_is_implicit_pooling():
            self.skipTest("sessions can change with implicit pooling")
        self.parse_count_info = ParseCountInfo()
        self.parse_count_info._initialize(conn or self.conn)

    def setup_round_trip_checker(self, conn=None):
        if get_is_implicit_pooling():
            self.skipTest("sessions can change with implicit pooling")
        self.round_trip_info = RoundTripInfo()
        self.round_trip_info._initialize(conn or self.conn)

    def tearDown(self):
        if self.requires_connection:
            self.conn.close()
            del self.cursor
            del self.conn


class BaseAsyncTestCase(unittest.IsolatedAsyncioTestCase):
    requires_connection = True

    async def assertParseCount(self, n):
        self.assertEqual(await self.parse_count_info.get_value_async(), n)

    def assertRaisesFullCode(self, *full_codes):
        return FullCodeErrorContextManager(full_codes)

    async def assertRoundTrips(self, n):
        self.assertEqual(await self.round_trip_info.get_value_async(), n)

    async def asyncSetUp(self):
        if self.requires_connection:
            self.conn = await get_connection_async()
            self.cursor = self.conn.cursor()
            await self.cursor.execute("alter session set time_zone = '+00:00'")

    async def asyncTearDown(self):
        if self.requires_connection:
            await self.conn.close()
            del self.cursor
            del self.conn

    async def get_db_object_as_plain_object(self, obj):
        if obj.type.iscollection:
            element_values = []
            for value in obj.aslist():
                if isinstance(value, oracledb.DbObject):
                    value = await self.get_db_object_as_plain_object(value)
                elif isinstance(value, oracledb.AsyncLOB):
                    value = await value.read()
                element_values.append(value)
            return element_values
        attr_values = []
        for attribute in obj.type.attributes:
            value = getattr(obj, attribute.name)
            if isinstance(value, oracledb.DbObject):
                value = await self.get_db_object_as_plain_object(value)
            elif isinstance(value, oracledb.AsyncLOB):
                value = await value.read()
            attr_values.append(value)
        return tuple(attr_values)

    async def get_sid_serial(self, conn=None):
        """
        Returns the sid and serial number of the connection as a 2-tuple.
        """
        if conn is None:
            conn = self.conn
        with conn.cursor() as cursor:
            await cursor.execute(
                """
                select
                    dbms_debug_jdwp.current_session_id,
                    dbms_debug_jdwp.current_session_serial
                from dual
                """
            )
            return await cursor.fetchone()

    async def is_on_oracle_cloud(self, connection=None):
        if connection is None:
            connection = self.conn
        return await is_on_oracle_cloud_async(connection)

    async def setup_parse_count_checker(self, conn=None):
        if get_is_implicit_pooling():
            self.skipTest("sessions can change with implicit pooling")
        self.parse_count_info = ParseCountInfo()
        await self.parse_count_info._initialize_async(conn or self.conn)

    async def setup_round_trip_checker(self, conn=None):
        if get_is_implicit_pooling():
            self.skipTest("sessions can change with implicit pooling")
        self.round_trip_info = RoundTripInfo()
        await self.round_trip_info._initialize_async(conn or self.conn)



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
2200 - Module for testing number variables
"""

import decimal

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def output_type_handler_binary_int(self, cursor, metadata):
        return cursor.var(
            oracledb.DB_TYPE_BINARY_INTEGER, arraysize=cursor.arraysize
        )

    def output_type_handler_decimal(self, cursor, metadata):
        if metadata.type_code is oracledb.DB_TYPE_NUMBER:
            return cursor.var(
                str,
                255,
                outconverter=decimal.Decimal,
                arraysize=cursor.arraysize,
            )

    def output_type_handler_str(self, cursor, metadata):
        return cursor.var(str, 255, arraysize=cursor.arraysize)

    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            number_col = i + i * 0.25
            float_col = i + i * 0.75
            unconstrained_col = i**3 + i * 0.5
            if i % 2:
                nullable_col = 143**i
            else:
                nullable_col = None
            data_tuple = (
                i,
                38**i,
                number_col,
                float_col,
                unconstrained_col,
                nullable_col,
            )
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_2200(self):
        "2200 - test binding in a boolean"
        result = self.cursor.callfunc(
            "pkg_TestBooleans.GetStringRep", str, [True]
        )
        self.assertEqual(result, "TRUE")

    def test_2201(self):
        "2201 - test binding in a boolean as a number"
        var = self.cursor.var(oracledb.NUMBER)
        var.setvalue(0, True)
        self.cursor.execute("select :1 from dual", [var])
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, 1)
        var.setvalue(0, False)
        self.cursor.execute("select :1 from dual", [var])
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, 0)

    def test_2202(self):
        "2202 - test binding in a decimal.Decimal"
        self.cursor.execute(
            """
            select *
            from TestNumbers
            where NumberCol - :value1 - :value2 = trunc(NumberCol)
            """,
            value1=decimal.Decimal("0.20"),
            value2=decimal.Decimal("0.05"),
        )
        expected_data = [
            self.data_by_key[1],
            self.data_by_key[5],
            self.data_by_key[9],
        ]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_2203(self):
        "2203 - test binding in a float"
        self.cursor.execute(
            """
            select *
            from TestNumbers
            where NumberCol - :value = trunc(NumberCol)
            """,
            value=0.25,
        )
        expected_data = [
            self.data_by_key[1],
            self.data_by_key[5],
            self.data_by_key[9],
        ]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_2204(self):
        "2204 - test binding in an integer"
        self.cursor.execute(
            "select * from TestNumbers where IntCol = :value",
            value=2,
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[2]])

    def test_2205(self):
        "2205 - test binding in a large long integer as Oracle number"
        in_val = 6088343244
        value_var = self.cursor.var(oracledb.NUMBER)
        value_var.setvalue(0, in_val)
        self.cursor.execute(
            """
            begin
                :value := :value + 5;
            end;
            """,
            value=value_var,
        )
        self.assertEqual(value_var.getvalue(), in_val + 5)

    def test_2206(self):
        "2206 - test binding in a large long integer as Python integer"
        long_value = -9999999999999999999
        self.cursor.execute("select :value from dual", value=long_value)
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, long_value)

    def test_2207(self):
        "2207 - test binding in an integer after setting input sizes to string"
        self.cursor.setinputsizes(value=15)
        self.cursor.execute(
            "select * from TestNumbers where IntCol = :value",
            value=3,
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[3]])

    def test_2208(self):
        "2208 - test binding in a decimal after setting input sizes to number"
        cursor = self.conn.cursor()
        value = decimal.Decimal("319438950232418390.273596")
        cursor.setinputsizes(value=oracledb.NUMBER)
        cursor.outputtypehandler = self.output_type_handler_decimal
        cursor.execute("select :value from dual", value=value)
        (out_value,) = cursor.fetchone()
        self.assertEqual(out_value, value)

    def test_2209(self):
        "2209 - test binding in a null"
        self.cursor.execute(
            "select * from TestNumbers where IntCol = :value",
            value=None,
        )
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2210(self):
        "2210 - test binding in a number array"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = [r[2] for r in self.raw_data]
        statement = """
                begin
                    :return_value := pkg_TestNumberArrays.TestInArrays(
                        :start_value, :array);
                end;"""
        self.cursor.execute(
            statement, return_value=return_value, start_value=5, array=array
        )
        self.assertEqual(return_value.getvalue(), 73.75)
        array = list(range(15))
        self.cursor.execute(statement, start_value=10, array=array)
        self.assertEqual(return_value.getvalue(), 115.0)

    def test_2211(self):
        "2211 - test binding in a number array (with setinputsizes)"
        return_value = self.cursor.var(oracledb.NUMBER)
        self.cursor.setinputsizes(array=[oracledb.NUMBER, 10])
        array = [r[2] for r in self.raw_data]
        self.cursor.execute(
            """
            begin
                :return_value := pkg_TestNumberArrays.TestInArrays(
                    :start_value, :array);
            end;
            """,
            return_value=return_value,
            start_value=6,
            array=array,
        )
        self.assertEqual(return_value.getvalue(), 74.75)

    def test_2212(self):
        "2212 - test binding in a number array (with arrayvar)"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = self.cursor.arrayvar(
            oracledb.NUMBER, [r[2] for r in self.raw_data]
        )
        self.cursor.execute(
            """
            begin
                :return_value := pkg_TestNumberArrays.TestInArrays(
                    :integer_value, :array);
            end;
            """,
            return_value=return_value,
            integer_value=7,
            array=array,
        )
        self.assertEqual(return_value.getvalue(), 75.75)

    def test_2213(self):
        "2213 - test binding in a zero length number array (with arrayvar)"
        return_value = self.cursor.var(oracledb.NUMBER)
        array = self.cursor.arrayvar(oracledb.NUMBER, 0)
        self.cursor.execute(
            """
            begin
                :return_value := pkg_TestNumberArrays.TestInArrays(
                    :integer_value, :array);
            end;
            """,
            return_value=return_value,
            integer_value=8,
            array=array,
        )
        self.assertEqual(return_value.getvalue(), 8.0)
        self.assertEqual(array.getvalue(), [])

    def test_2214(self):
        "2214 - test binding in/out a number array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.NUMBER, 10)
        original_data = [r[2] for r in self.raw_data]
        expected_data = [
            original_data[i - 1] * 10 for i in range(1, 6)
        ] + original_data[5:]
        array.setvalue(0, original_data)
        self.cursor.execute(
            """
            begin
                pkg_TestNumberArrays.TestInOutArrays(:num_elems, :array);
            end;
            """,
            num_elems=5,
            array=array,
        )
        self.assertEqual(array.getvalue(), expected_data)

    def test_2215(self):
        "2215 - test binding out a Number array (with arrayvar)"
        array = self.cursor.arrayvar(oracledb.NUMBER, 6)
        expected_data = [i * 100 for i in range(1, 7)]
        self.cursor.execute(
            """
            begin
                pkg_TestNumberArrays.TestOutArrays(:num_elems, :array);
            end;
            """,
            num_elems=6,
            array=array,
        )
        self.assertEqual(array.getvalue(), expected_data)

    def test_2216(self):
        "2216 - test binding out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.NUMBER)
        self.cursor.execute(
            """
            begin
                :value := 5;
            end;
            """
        )
        self.assertEqual(bind_vars["value"].getvalue(), 5)

    def test_2217(self):
        "2217 - test binding in/out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.NUMBER)
        self.cursor.execute(
            """
            begin
                :value := :value + 5;
            end;
            """,
            value=1.25,
        )
        self.assertEqual(bind_vars["value"].getvalue(), 6.25)

    def test_2218(self):
        "2218 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.NUMBER)
        self.cursor.execute(
            """
            begin
                :value := 5;
            end;
            """,
            value=var,
        )
        self.assertEqual(var.getvalue(), 5)

    def test_2219(self):
        "2219 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.NUMBER)
        var.setvalue(0, 2.25)
        self.cursor.execute(
            """
            begin
                :value := :value + 5;
            end;
            """,
            value=var,
        )
        self.assertEqual(var.getvalue(), 7.25)

    def test_2220(self):
        "2220 - test cursor description is accurate"
        self.cursor.execute("select * from TestNumbers")
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ("LONGINTCOL", oracledb.DB_TYPE_NUMBER, 17, None, 16, 0, False),
            ("NUMBERCOL", oracledb.DB_TYPE_NUMBER, 13, None, 9, 2, False),
            ("FLOATCOL", oracledb.DB_TYPE_NUMBER, 127, None, 126, -127, False),
            (
                "UNCONSTRAINEDCOL",
                oracledb.DB_TYPE_NUMBER,
                127,
                None,
                0,
                -127,
                False,
            ),
            ("NULLABLECOL", oracledb.DB_TYPE_NUMBER, 39, None, 38, 0, True),
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_2221(self):
        "2221 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestNumbers order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2222(self):
        "2222 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestNumbers order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_2223(self):
        "2223 - test that fetching a single row returns the correct results"
        self.cursor.execute(
            """
            select *
            from TestNumbers
            where IntCol in (3, 4)
            order by IntCol
            """
        )
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertIsNone(self.cursor.fetchone())

    def test_2224(self):
        "2224 - test that fetching a long integer returns such in Python"
        self.cursor.execute(
            """
            select NullableCol
            from TestNumbers
            where IntCol = 9
            """
        )
        (col,) = self.cursor.fetchone()
        self.assertEqual(col, 25004854810776297743)

    def test_2225(self):
        "2225 - test fetching a floating point number returns such in Python"
        self.cursor.execute("select 1.25 from dual")
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, 1.25)

    def test_2226(self):
        "2226 - test that fetching an integer returns such in Python"
        self.cursor.execute("select 148 from dual")
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, 148)
        self.assertIsInstance(result, int, "integer not returned")

    def test_2227(self):
        "2227 - test that acceptable boundary numbers are handled properly"
        in_values = [
            decimal.Decimal("9.99999999999999e+125"),
            decimal.Decimal("-9.99999999999999e+125"),
            0.0,
            1e-130,
            -1e-130,
        ]
        out_values = [
            int("9" * 15 + "0" * 111),
            -int("9" * 15 + "0" * 111),
            0,
            1e-130,
            -1e-130,
        ]
        for in_value, out_value in zip(in_values, out_values):
            self.cursor.execute("select :1 from dual", [in_value])
            (result,) = self.cursor.fetchone()
            self.assertEqual(result, out_value)

    def test_2228(self):
        "2228 - test that unacceptable boundary numbers are rejected"
        test_values = [
            (1e126, "DPY-4003"),
            (-1e126, "DPY-4003"),
            (float("inf"), "DPY-4004"),
            (float("-inf"), "DPY-4004"),
            (float("NaN"), "DPY-4004"),
            (decimal.Decimal("1e126"), "DPY-4003"),
            (decimal.Decimal("-1e126"), "DPY-4003"),
            (decimal.Decimal("inf"), "DPY-4004"),
            (decimal.Decimal("-inf"), "DPY-4004"),
            (decimal.Decimal("NaN"), "DPY-4004"),
        ]
        for value, error in test_values:
            with self.assertRaisesFullCode(error):
                self.cursor.execute("select :1 from dual", [value])

    def test_2229(self):
        "2229 - test that fetching the result of division returns a float"
        self.cursor.execute(
            """
            select IntCol / 7
            from TestNumbers
            where IntCol = 1
            """
        )
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, 1.0 / 7.0)
        self.assertIsInstance(result, float, "float not returned")

    def test_2230(self):
        "2230 - test that string format is returned properly"
        var = self.cursor.var(oracledb.NUMBER)
        self.assertIs(var.type, oracledb.DB_TYPE_NUMBER)
        self.assertEqual(
            str(var), "<oracledb.Var of type DB_TYPE_NUMBER with value None>"
        )
        var.setvalue(0, 4.5)
        self.assertEqual(
            str(var), "<oracledb.Var of type DB_TYPE_NUMBER with value 4.5>"
        )

    def test_2231(self):
        "2231 - test that binding binary double is possible"
        statement = "select :1 from dual"
        self.cursor.setinputsizes(oracledb.DB_TYPE_BINARY_DOUBLE)
        self.cursor.execute(statement, (5,))
        self.assertEqual(
            self.cursor.bindvars[0].type, oracledb.DB_TYPE_BINARY_DOUBLE
        )
        (value,) = self.cursor.fetchone()
        self.assertEqual(value, 5)

        self.cursor.execute(statement, (1.5,))
        self.assertEqual(
            self.cursor.bindvars[0].type, oracledb.DB_TYPE_BINARY_DOUBLE
        )
        (value,) = self.cursor.fetchone()
        self.assertEqual(value, 1.5)

        self.cursor.execute(statement, [decimal.Decimal("NaN")])
        self.assertEqual(
            self.cursor.bindvars[0].type, oracledb.DB_TYPE_BINARY_DOUBLE
        )
        (value,) = self.cursor.fetchone()
        self.assertEqual(str(value), str(float("NaN")))

    def test_2232(self):
        "2232 - test fetching numbers as binary integers"
        self.cursor.outputtypehandler = self.output_type_handler_binary_int
        for value in (1, 2**31, 2**63 - 1, -1, -(2**31), -(2**63) + 1):
            self.cursor.execute("select :1 from dual", [str(value)])
            (fetched_value,) = self.cursor.fetchone()
            self.assertEqual(value, fetched_value)

    def test_2233(self):
        "2233 - test binding native integer as an out bind"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        self.cursor.execute("begin :value := 2.9; end;", [simple_var])
        self.assertEqual(simple_var.getvalue(), 2)

        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        self.cursor.execute("begin :value := 1.5; end;", [simple_var])
        self.assertEqual(simple_var.getvalue(), 1)

    def test_2234(self):
        "2234 - test binding in a native integer"
        statement = "begin :value := :value + 2.5; end;"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        simple_var.setvalue(0, 0)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), 2)

        simple_var.setvalue(0, -5)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), -2)

    def test_2235(self):
        "2235 - test setting decimal value for binary int"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        simple_var.setvalue(0, 2.5)
        self.cursor.execute("begin :value := :value + 2.5; end;", [simple_var])
        self.assertEqual(simple_var.getvalue(), 4)

    def test_2236(self):
        "2236 - bind a large value to binary int"
        simple_var = self.cursor.var(oracledb.DB_TYPE_BINARY_INTEGER)
        self.cursor.execute(
            "begin :value := POWER(2, 31) - 1; end;", [simple_var]
        )
        self.assertEqual(simple_var.getvalue(), 2**31 - 1)

        self.cursor.execute(
            "begin :value := POWER(-2, 31) - 1; end;", [simple_var]
        )
        self.assertEqual(simple_var.getvalue(), -(2**31) - 1)

    def test_2237(self):
        "2237 - fetch a number with oracledb.defaults.fetch_lobs = False"
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute("select 1 from dual")
            (result,) = self.cursor.fetchone()
            self.assertIsInstance(result, int)

    def test_2238(self):
        "2238 - fetch a small constant with a decimal point"
        self.cursor.outputtypehandler = self.output_type_handler_str
        self.cursor.execute("select 3 / 2 from dual")
        (result,) = self.cursor.fetchone()
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], "1")
        self.assertEqual(result[2], "5")


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1000 - Module for testing pool strinking. No special setup is required but the
tests here will only be run if the run_long_tests value is enabled.
"""

import time
import unittest

import test_env


@unittest.skipUnless(
    test_env.get_extended_config_bool("run_long_tests"),
    "extended configuration run_long_tests is disabled",
)
class TestCase(test_env.BaseTestCase):
    def test_ext_1000(self):
        "E1000 - test pool timeout with simple acquire after waiting"
        pool = test_env.get_pool(min=3, max=10, increment=1, timeout=5)
        conns = [pool.acquire() for i in range(7)]
        self.assertEqual(pool.opened, 7)
        for conn in conns:
            conn.close()
        time.sleep(10)
        conn = pool.acquire()
        self.assertEqual(pool.opened, 3)

    def test_ext_1001(self):
        "E1001 - test pool timeout with older connection returned first"
        pool = test_env.get_pool(min=2, max=5, increment=1, timeout=3)
        conns = [pool.acquire() for i in range(3)]
        conns[2].close()
        for i in range(10):
            with pool.acquire() as conn:
                with conn.cursor() as cursor:
                    cursor.execute("select 1 from dual")
        time.sleep(4)
        conn = pool.acquire()
        self.assertEqual(pool.opened, 3)

    @unittest.skipIf(not test_env.get_is_thin(), "doesn't occur in thick mode")
    def test_ext_1002(self):
        "E1002 - test pool shrinks to min on pool inactivity"
        pool = test_env.get_pool(min=3, max=10, increment=2, timeout=4)
        conns = [pool.acquire() for i in range(6)]
        self.assertEqual(pool.opened, 6)
        for conn in conns:
            conn.close()
        time.sleep(6)
        self.assertEqual(pool.opened, 3)

    @unittest.skipIf(not test_env.get_is_thin(), "doesn't occur in thick mode")
    def test_ext_1003(self):
        "E1003 - test pool eliminates extra connections on inactivity"
        pool = test_env.get_pool(min=4, max=10, increment=4, timeout=3)
        conns = [pool.acquire() for i in range(5)]
        self.assertEqual(pool.opened, 5)
        time.sleep(2)
        self.assertEqual(pool.opened, 8)
        time.sleep(3)
        self.assertEqual(pool.opened, 5)
        del conns


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1700 - Module for testing the generation of warnings with asyncio. No special
setup is required but sys privilege is required in order to create and drop
users.
"""

import asyncio
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    profile_name = "profile_priv_test_1700"
    user_name = "user_priv_test_1700"
    requires_connection = False
    setup_completed = False

    async def __perform_setup(self):
        """
        Perform the setup, if necessary.
        """
        if self.__class__.setup_completed:
            return
        conn = await test_env.get_admin_connection_async()
        password = test_env.get_main_password()
        cursor = conn.cursor()
        await cursor.execute(
            f"""
            declare
                e_user_missing exception;
                pragma exception_init(e_user_missing, -1918);
            begin
                execute immediate('drop user {self.user_name} cascade');
            exception
            when e_user_missing then
                null;
            end;
            """
        )
        await cursor.execute(
            f"""
            declare
                e_user_missing exception;
                pragma exception_init(e_user_missing, -2380);
            begin
                execute immediate('drop profile {self.profile_name}');
            exception
            when e_user_missing then
                null;
            end;
            """
        )
        await cursor.execute(
            f"create user {self.user_name} identified by {password}"
        )
        await cursor.execute(f"grant create session to {self.user_name}")
        await cursor.execute(
            f"""
            create profile {self.profile_name} limit
            password_life_time 1 / 24 / 60 / 60
            password_grace_time 7
            """
        )
        await cursor.execute(
            f"alter user {self.user_name} profile {self.profile_name}"
        )
        await asyncio.sleep(2)
        self.__class__.setup_completed = True

    async def test_ext_1700(self):
        "E1700 - test standalone connection generates a warning"
        await self.__perform_setup()
        password = test_env.get_main_password()
        async with oracledb.connect_async(
            user=self.user_name,
            password=password,
            dsn=test_env.get_connect_string(),
        ) as conn:
            self.assertIn(conn.warning.full_code, ["ORA-28002", "ORA-28098"])

    async def test_ext_1701(self):
        "E1701 - test pooled connection generates a warning (min 0)"
        await self.__perform_setup()
        password = test_env.get_main_password()
        pool = oracledb.create_pool_async(
            user=self.user_name,
            password=password,
            dsn=test_env.get_connect_string(),
            min=0,
            max=5,
            increment=1,
        )
        async with pool.acquire() as conn:
            self.assertIn(conn.warning.full_code, ["ORA-28002", "ORA-28098"])
        async with pool.acquire() as conn:
            self.assertIsNone(conn.warning)
        await pool.close(0)

    async def test_ext_1702(self):
        "E1702 - test pooled connection generates a warning (min 1)"
        await self.__perform_setup()
        password = test_env.get_main_password()
        pool = oracledb.create_pool_async(
            user=self.user_name,
            password=password,
            dsn=test_env.get_connect_string(),
            min=1,
            max=5,
            increment=1,
        )
        async with pool.acquire() as conn:
            self.assertIn(conn.warning.full_code, ["ORA-28002", "ORA-28098"])
        async with pool.acquire() as conn:
            self.assertIsNone(conn.warning)
        await pool.close(0)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# protocol.pyx
#
# Cython file defining the protocol used by the client when communicating with
# the database (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseProtocol:

    cdef:
        uint8_t _seq_num
        Transport _transport
        Capabilities _caps
        ReadBuffer _read_buf
        WriteBuffer _write_buf
        bint _in_connect
        bint _txn_in_progress
        bint _break_in_progress
        object _request_lock

    def __init__(self):
        self._caps = Capabilities()
        # mark protocol to indicate that connect is in progress; this prevents
        # the normal break/reset mechanism from firing, which is unnecessary
        # since the connection is going to be immediately closed anyway!
        self._in_connect = True
        self._transport = Transport.__new__(Transport)
        self._transport._max_packet_size = self._caps.sdu
        self._read_buf = ReadBuffer(self._transport, self._caps)
        self._write_buf = WriteBuffer(self._transport, self._caps)

    cdef int _break_external(self) except -1:
        """
        Method for sending a break to the server from an external request. A
        separate write buffer is used in order to avoid a potential conflict
        with any in progress writes.
        """
        cdef WriteBuffer buf
        if not self._break_in_progress:
            self._break_in_progress = True
            if self._caps.supports_oob:
                self._transport.send_oob_break()
            else:
                buf = WriteBuffer(self._transport, self._caps)
                self._send_marker(buf, TNS_MARKER_TYPE_INTERRUPT)

    cdef int _final_close(self, WriteBuffer buf) except -1:
        """
        Send the final close packet to the server and close the socket.
        """
        buf.start_request(TNS_PACKET_TYPE_DATA, 0, TNS_DATA_FLAGS_EOF)
        buf.end_request()
        self._force_close()

    cdef int _force_close(self) except -1:
        """
        Forces the connection closed. This is used when an unrecoverable error
        has taken place.
        """
        cdef Transport transport = self._transport
        if transport is not None:
            self._transport = None
            self._read_buf._transport = None
            self._write_buf._transport = None
            transport.disconnect()

    cdef int _post_connect(self, BaseThinConnImpl conn_impl,
                           AuthMessage auth_message) except -1:
        """"
        Performs activities after the connection has completed. The protocol
        must be marked to indicate that the connect is no longer in progress,
        which allows the normal break/reset mechanism to fire. The session must
        also be marked as not needing to be closed since for listener redirects
        the packet may indicate EOF for the initial connection that is
        established.
        """
        conn_impl.warning = auth_message.warning
        self._read_buf._pending_error_num = 0
        self._in_connect = False

    cdef int _release_drcp_session(self, BaseThinConnImpl conn_impl,
                                   uint32_t release_mode) except -1:
        """
        Release the session back to DRCP. Standalone sessions are marked for
        deauthentication.
        """
        cdef SessionReleaseMessage message
        message = conn_impl._create_message(SessionReleaseMessage)
        message.release_mode = release_mode
        message.send(self._write_buf)

    cdef int _send_marker(self, WriteBuffer buf, uint8_t marker_type):
        """
        Sends a marker of the specified type to the server.
        Internal method for sending a break to the server.
        """
        buf.start_request(TNS_PACKET_TYPE_MARKER)
        buf.write_uint8(1)
        buf.write_uint8(0)
        buf.write_uint8(marker_type)
        buf.end_request()

    cdef int _process_call_status(self, BaseThinConnImpl conn_impl,
                                  uint32_t call_status) except -1:
        """
        Processes the call status flags returned by the server.
        """
        self._txn_in_progress = call_status & TNS_EOCS_FLAGS_TXN_IN_PROGRESS
        if call_status & TNS_EOCS_FLAGS_SESS_RELEASE:
            conn_impl._statement_cache.clear_open_cursors()


cdef class Protocol(BaseProtocol):

    def __init__(self):
        BaseProtocol.__init__(self)
        self._request_lock = threading.Lock()

    cdef int _close(self, ThinConnImpl conn_impl) except -1:
        """
        Closes the connection. If a transaction is in progress it will be
        rolled back. DRCP sessions will be released. For standalone
        connections, the session will be logged off. For pooled connections,
        the connection will be returned to the pool for subsequent use.
        """
        cdef:
            uint32_t release_mode = DRCP_DEAUTHENTICATE \
                    if conn_impl._pool is None else 0
            ThinPoolImpl pool_impl
            Message message

        with self._request_lock:

            # if a read failed on the socket earlier, clear the socket
            if self._read_buf._transport is None \
                    or self._read_buf._transport._transport is None:
                self._transport = None

            # if the session was marked as needing to be closed, force it
            # closed immediately (unless it was already closed)
            if self._read_buf._pending_error_num != 0 \
                    and self._transport is not None:
                self._force_close()

            # rollback any open transaction and release the DRCP session, if
            # applicable
            if self._transport is not None:
                if self._txn_in_progress:
                    if conn_impl._transaction_context is not None:
                        message = conn_impl._create_tpc_rollback_message()
                    else:
                        message = conn_impl._create_message(RollbackMessage)
                    self._process_message(message)
                    conn_impl._transaction_context = None
                if conn_impl._drcp_enabled:
                    self._release_drcp_session(conn_impl, release_mode)
                    conn_impl._drcp_establish_session = True

            # if the connection is part of a pool, return it to the pool
            if conn_impl._pool is not None:
                pool_impl = <ThinPoolImpl> conn_impl._pool
                return pool_impl._return_connection(conn_impl)

            # otherwise, destroy the database object type cache, send the
            # logoff message and final close packet
            if conn_impl._dbobject_type_cache_num > 0:
                remove_dbobject_type_cache(conn_impl._dbobject_type_cache_num)
                conn_impl._dbobject_type_cache_num = 0
            if self._transport is not None:
                if not conn_impl._drcp_enabled:
                    message = conn_impl._create_message(LogoffMessage)
                    self._process_message(message)
                self._final_close(self._write_buf)

    cdef int _connect_phase_one(self, ThinConnImpl conn_impl,
                                ConnectParamsImpl params,
                                Description description,
                                Address address,
                                str connect_string) except -1:
        """
        Method for performing the required steps for establishing a connection
        within the scope of a retry. If the listener refuses the connection, a
        retry will be performed, if retry_count is set.
        """
        cdef:
            ConnectMessage connect_message = None
            uint8_t packet_type, packet_flags = 0
            object ssl_context, connect_info
            ConnectParamsImpl temp_params
            str host, redirect_data
            Address temp_address
            int port, pos

        # store whether OOB processing is possible or not
        self._caps.supports_oob = not params.disable_oob \
                and sys.platform != "win32"

        # establish initial TCP connection and get initial connect string
        host = address.host
        port = address.port
        self._connect_tcp(params, description, address, host, port,
                          connect_string)

        # send connect message and process response; this may request the
        # message to be resent multiple times; if a redirect packet is
        # detected, a new TCP connection is established first
        while True:

            # create connection message, if needed
            if connect_message is None:
                connect_message = conn_impl._create_message(ConnectMessage)
                connect_message.host = host
                connect_message.port = port
                connect_message.description = description
                connect_message.connect_string_bytes = connect_string.encode()
                connect_message.connect_string_len = \
                        <uint16_t> len(connect_message.connect_string_bytes)
                connect_message.packet_flags = packet_flags

            # process connection message
            self._process_message(connect_message)
            packet_type = self._read_buf._current_packet.packet_type
            if connect_message.redirect_data is not None:
                redirect_data = connect_message.redirect_data
                pos = redirect_data.find('\x00')
                if pos < 0:
                    errors._raise_err(errors.ERR_INVALID_REDIRECT_DATA,
                                      data=redirect_data)
                temp_params = ConnectParamsImpl()
                temp_params._parse_connect_string(redirect_data[:pos])
                temp_address = temp_params._get_addresses()[0]
                host = temp_address.host
                port = temp_address.port
                connect_string = redirect_data[pos + 1:]
                self._connect_tcp(params, description, address, host, port,
                                  connect_string)
                connect_message = None
                packet_flags = TNS_PACKET_FLAG_REDIRECT
            elif packet_type == TNS_PACKET_TYPE_ACCEPT:
                self._transport._max_packet_size = self._caps.sdu
                self._write_buf._size_for_sdu()
                break

            # for TCPS connections, OOB processing is not supported; if the
            # packet flags indicate that TLS renegotiation is required, this is
            # performed now
            if address.protocol == "tcps":
                self._caps.supports_oob = False
                packet_flags = self._read_buf._current_packet.packet_flags
                if packet_flags & TNS_PACKET_FLAG_TLS_RENEG:
                    self._transport.renegotiate_tls(description)

    cdef int _connect_phase_two(self, ThinConnImpl conn_impl,
                                Description description,
                                ConnectParamsImpl params) except -1:
        """"
        Method for perfoming the required steps for establishing a connection
        oustide the scope of a retry. If any of the steps in this method fail,
        an exception will be raised.
        """
        cdef:
            DataTypesMessage data_types_message
            FastAuthMessage fast_auth_message
            ProtocolMessage protocol_message
            bint supports_end_of_response
            AuthMessage auth_message

        # if we can use OOB, send an urgent message now followed by a reset
        # marker to see if the server understands it
        if self._caps.supports_oob \
                and self._caps.protocol_version >= TNS_VERSION_MIN_OOB_CHECK:
            self._transport.send_oob_break()
            self._send_marker(self._write_buf, TNS_MARKER_TYPE_RESET)

        # create the messages that need to be sent to the server
        protocol_message = conn_impl._create_message(ProtocolMessage)
        data_types_message = conn_impl._create_message(DataTypesMessage)
        auth_message = conn_impl._create_message(AuthMessage)
        auth_message._set_params(params, description)

        # starting in 23ai, fast authentication is possible; see if the server
        # supports it
        if self._caps.supports_fast_auth:
            fast_auth_message = conn_impl._create_message(FastAuthMessage)
            fast_auth_message.protocol_message = protocol_message
            fast_auth_message.data_types_message = data_types_message
            fast_auth_message.auth_message = auth_message
            self._process_message(fast_auth_message)

        # otherwise, do the normal authentication; disable end of response for
        # the first two messages as the server does not send an end of response
        # for these messages
        else:
            supports_end_of_response = self._caps.supports_end_of_response
            self._caps.supports_end_of_response = False
            self._process_message(protocol_message)
            self._process_message(data_types_message)
            self._caps.supports_end_of_response = supports_end_of_response
            self._process_message(auth_message)

        # send authorization message a second time, if needed, to respond to
        # the challenge sent by the server
        if auth_message.resend:
            self._process_message(auth_message)

        # perform post connect activities
        self._post_connect(conn_impl, auth_message)

    cdef int _connect_tcp(self, ConnectParamsImpl params,
                          Description description, Address address, str host,
                          int port, str connect_string) except -1:
        """
        Creates a socket on which to communicate using the provided parameters.
        If a proxy is configured, a connection to the proxy is established and
        the target host and port is forwarded to the proxy.
        """
        cdef:
            bint use_proxy = (address.https_proxy is not None)
            double timeout = description.tcp_connect_timeout
            bint use_tcps = (address.protocol == "tcps")
            object connect_info, sock, data, reply, m

        # establish connection to appropriate host/port
        if use_proxy:
            if not use_tcps:
                errors._raise_err(errors.ERR_HTTPS_PROXY_REQUIRES_TCPS)
            connect_info = (address.https_proxy, address.https_proxy_port)
        else:
            connect_info = (host, port)
            if not use_tcps and (params._token is not None
                    or params.access_token_callback is not None):
                errors._raise_err(errors.ERR_ACCESS_TOKEN_REQUIRES_TCPS)
        if description.use_tcp_fast_open:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.sendto(connect_string.encode(), socket.MSG_FASTOPEN,
                        connect_info)
        else:
            sock = socket.create_connection(connect_info, timeout)

        # complete connection through proxy, if applicable
        if use_proxy:
            data = f"CONNECT {host}:{port} HTTP/1.0\r\n\r\n"
            sock.send(data.encode())
            reply = sock.recv(1024)
            m = re.search('HTTP/1.[01]\\s+(\\d+)\\s+', reply.decode())
            if m is None or m.groups()[0] != '200':
                errors._raise_err(errors.ERR_PROXY_FAILURE,
                                  response=reply.decode())

        # set socket on transport
        self._transport.set_from_socket(sock, params, description, address)

        # negotiate TLS, if applicable
        if use_tcps:
            self._transport.create_ssl_context(params, description, address)
            self._transport.negotiate_tls(sock, description)

    cdef int _process_message(self, Message message) except -1:
        cdef uint32_t timeout = message.conn_impl._call_timeout
        try:
            self._read_buf.reset_packets()
            message.send(self._write_buf)
            self._receive_packet(message, check_request_boundary=True)
            message.process(self._read_buf)
        except socket.timeout:
            try:
                self._break_external()
                self._receive_packet(message)
                self._break_in_progress = False
                errors._raise_err(errors.ERR_CALL_TIMEOUT_EXCEEDED,
                                  timeout=timeout)
            except socket.timeout:
                self._force_close()
                errors._raise_err(errors.ERR_CONNECTION_CLOSED,
                                  "socket timed out while recovering from " \
                                  "previous socket timeout")
            raise
        except Exception as e:
            if not self._in_connect \
                    and self._write_buf._packet_sent \
                    and self._read_buf._transport is not None \
                    and self._read_buf._transport._transport is not None:
                self._send_marker(self._write_buf, TNS_MARKER_TYPE_BREAK)
                self._reset(message)
            raise
        if message.flush_out_binds:
            self._write_buf.start_request(TNS_PACKET_TYPE_DATA)
            self._write_buf.write_uint8(TNS_MSG_TYPE_FLUSH_OUT_BINDS)
            self._write_buf.end_request()
            self._receive_packet(message)
            message.process(self._read_buf)
        if self._break_in_progress:
            try:
                if self._caps.supports_oob:
                    self._send_marker(self._write_buf,
                                      TNS_MARKER_TYPE_INTERRUPT)
                self._receive_packet(message)
            except socket.timeout:
                errors._raise_err(errors.ERR_CONNECTION_CLOSED,
                                  "socket timed out while awaiting break " \
                                  "response from server")
            message.process(self._read_buf)
            self._break_in_progress = False
        self._process_call_status(message.conn_impl, message.call_status)
        if message.error_occurred:
            if message.retry:
                message.error_occurred = False
                return self._process_message(message)
            message._check_and_raise_exception()

    cdef int _process_single_message(self, Message message) except -1:
        """
        Process a single message within a request.
        """
        message.preprocess()
        with self._request_lock:
            self._process_message(message)
            if message.resend:
                self._process_message(message)
        message.postprocess()

    cdef int _receive_packet(self, Message message,
                             bint check_request_boundary=False) except -1:
        cdef:
            ReadBuffer buf = self._read_buf
            uint16_t refuse_message_len
            const char_type* ptr
        buf._check_request_boundary = \
                check_request_boundary and self._caps.supports_end_of_response
        buf.wait_for_packets_sync()
        buf._check_request_boundary = False
        if buf._current_packet.packet_type == TNS_PACKET_TYPE_MARKER:
            self._reset(message)
        elif buf._current_packet.packet_type == TNS_PACKET_TYPE_REFUSE:
            self._write_buf._packet_sent = False
            buf.skip_raw_bytes(2)
            buf.read_uint16(&refuse_message_len)
            if refuse_message_len == 0:
                message.error_info.message = None
            else:
                ptr = buf.read_raw_bytes(refuse_message_len)
                message.error_info.message = ptr[:refuse_message_len].decode()

    cdef int _reset(self, Message message) except -1:
        cdef uint8_t marker_type, packet_type

        # send reset marker
        self._send_marker(self._write_buf, TNS_MARKER_TYPE_RESET)

        # read and discard all packets until a reset marker is received
        while True:
            packet_type = self._read_buf._current_packet.packet_type
            if packet_type == TNS_PACKET_TYPE_MARKER:
                self._read_buf.skip_raw_bytes(2)
                self._read_buf.read_ub1(&marker_type)
                if marker_type == TNS_MARKER_TYPE_RESET:
                    break
            self._read_buf.wait_for_packets_sync()

        # read error packet; first skip as many marker packets as may be sent
        # by the server; if the server doesn't handle out-of-band breaks
        # properly, some quit immediately and others send multiple reset
        # markers (this addresses both situations without resulting in strange
        # errors)
        while packet_type == TNS_PACKET_TYPE_MARKER:
            self._read_buf.wait_for_packets_sync()
            packet_type = self._read_buf._current_packet.packet_type
        self._break_in_progress = False


cdef class BaseAsyncProtocol(BaseProtocol):

    def __init__(self):
        BaseProtocol.__init__(self)
        self._request_lock = asyncio.Lock()
        self._transport._is_async = True

    async def _close(self, AsyncThinConnImpl conn_impl):
        """
        Closes the connection. If a transaction is in progress it will be
        rolled back. DRCP sessions will be released. For standalone
        connections, the session will be logged off. For pooled connections,
        the connection will be returned to the pool for subsequent use.
        """
        cdef:
            uint32_t release_mode = DRCP_DEAUTHENTICATE \
                    if conn_impl._pool is None else 0
            AsyncThinPoolImpl pool_impl
            Message message

        async with self._request_lock:

            # if a read failed on the socket earlier, clear the socket
            if self._read_buf._transport is None:
                self._transport = None

            # if the session was marked as needing to be closed, force it
            # closed immediately (unless it was already closed)
            if self._read_buf._pending_error_num != 0 \
                    and self._transport is not None:
                self._force_close()

            # rollback any open transaction and release the DRCP session, if
            # applicable
            if self._transport is not None:
                if self._txn_in_progress:
                    message = conn_impl._create_message(RollbackMessage)
                    await self._process_message(message)
                if conn_impl._drcp_enabled:
                    self._release_drcp_session(conn_impl, release_mode)
                    conn_impl._drcp_establish_session = True

            # if the connection is part of a pool, return it to the pool
            if conn_impl._pool is not None:
                pool_impl = <AsyncThinPoolImpl> conn_impl._pool
                return await pool_impl._return_connection(conn_impl)

            # otherwise, destroy the database object type cache, send the
            # logoff message and final close packet
            if conn_impl._dbobject_type_cache_num > 0:
                remove_dbobject_type_cache(conn_impl._dbobject_type_cache_num)
                conn_impl._dbobject_type_cache_num = 0
            if self._transport is not None:
                if not conn_impl._drcp_enabled:
                    message = conn_impl._create_message(LogoffMessage)
                    await self._process_message(message)
                self._final_close(self._write_buf)

    async def _connect_phase_one(self,
                                 AsyncThinConnImpl conn_impl,
                                 ConnectParamsImpl params,
                                 Description description,
                                 Address address,
                                 str connect_string):
        """
        Method for performing the required steps for establishing a connection
        within the scope of a retry. If the listener refuses the connection, a
        retry will be performed, if retry_count is set.
        """
        cdef:
            ConnectMessage connect_message = None
            uint8_t packet_type, packet_flags = 0
            object ssl_context, connect_info
            ConnectParamsImpl temp_params
            str host, redirect_data
            object orig_transport
            Address temp_address
            int port, pos

        # asyncio doesn't support OOB processing
        self._caps.supports_oob = False

        # establish initial TCP connection and get initial connect string
        host = address.host
        port = address.port
        orig_transport = await self._connect_tcp(params, description, address,
                                                 host, port)

        # send connect message and process response; this may request the
        # message to be resent multiple times; if a redirect packet is
        # detected, a new TCP connection is established first
        while True:

            # create connection message, if needed
            if connect_message is None:
                connect_message = conn_impl._create_message(ConnectMessage)
                connect_message.host = host
                connect_message.port = port
                connect_message.description = description
                connect_message.connect_string_bytes = connect_string.encode()
                connect_message.connect_string_len = \
                        <uint16_t> len(connect_message.connect_string_bytes)
                connect_message.packet_flags = packet_flags

            # process connection message
            await self._process_message(connect_message)
            packet_type = self._read_buf._current_packet.packet_type
            if connect_message.redirect_data is not None:
                redirect_data = connect_message.redirect_data
                pos = redirect_data.find('\x00')
                if pos < 0:
                    errors._raise_err(errors.ERR_INVALID_REDIRECT_DATA,
                                      data=redirect_data)
                temp_params = ConnectParamsImpl()
                temp_params._parse_connect_string(redirect_data[:pos])
                temp_address = temp_params._get_addresses()[0]
                host = temp_address.host
                port = temp_address.port
                connect_string = redirect_data[pos + 1:]
                orig_transport = await self._connect_tcp(params, description,
                                                         address, host, port)
                connect_message = None
                packet_flags = TNS_PACKET_FLAG_REDIRECT
            elif packet_type == TNS_PACKET_TYPE_ACCEPT:
                self._transport._max_packet_size = self._caps.sdu
                self._write_buf._size_for_sdu()
                break

            # for TCPS connections, OOB processing is not supported; if the
            # packet flags indicate that TLS renegotiation is required, this is
            # performed now
            if address.protocol == "tcps":
                self._caps.supports_oob = False
                packet_flags = self._read_buf._current_packet.packet_flags
                if packet_flags & TNS_PACKET_FLAG_TLS_RENEG:
                    self._transport._transport = orig_transport
                    await self._transport.negotiate_tls_async(self,
                                                              description)

    async def _connect_phase_two(self, AsyncThinConnImpl conn_impl,
                                 Description description,
                                 ConnectParamsImpl params):
        """"
        Method for perfoming the required steps for establishing a connection
        oustide the scope of a retry. If any of the steps in this method fail,
        an exception will be raised.
        """
        cdef:
            DataTypesMessage data_types_message
            FastAuthMessage fast_auth_message
            ProtocolMessage protocol_message
            bint supports_end_of_response
            AuthMessage auth_message

        # create the messages that need to be sent to the server
        protocol_message = conn_impl._create_message(ProtocolMessage)
        data_types_message = conn_impl._create_message(DataTypesMessage)
        auth_message = conn_impl._create_message(AuthMessage)
        auth_message._set_params(params, description)

        # starting in 23ai, fast authentication is possible; see if the server
        # supports it
        if self._caps.supports_fast_auth:
            fast_auth_message = conn_impl._create_message(FastAuthMessage)
            fast_auth_message.protocol_message = protocol_message
            fast_auth_message.data_types_message = data_types_message
            fast_auth_message.auth_message = auth_message
            await self._process_message(fast_auth_message)

        # otherwise, do the normal authentication; disable end of response for
        # the first two messages as the server does not send an end of response
        # for these messages
        else:
            supports_end_of_response = self._caps.supports_end_of_response
            self._caps.supports_end_of_response = False
            await self._process_message(protocol_message)
            await self._process_message(data_types_message)
            self._caps.supports_end_of_response = supports_end_of_response
            await self._process_message(auth_message)

        # send authorization message a second time, if needed, to respond to
        # the challenge sent by the server
        if auth_message.resend:
            await self._process_message(auth_message)

        # perform post connect activities
        self._post_connect(conn_impl, auth_message)


    async def _connect_tcp(self, ConnectParamsImpl params,
                           Description description, Address address, str host,
                           int port):
        """
        Creates a socket on which to communicate using the provided parameters.
        If a proxy is configured, a connection to the proxy is established and
        the target host and port is forwarded to the proxy.
        """
        cdef:
            bint use_proxy = (address.https_proxy is not None)
            double timeout = description.tcp_connect_timeout
            bint use_tcps = (address.protocol == "tcps")
            object connect_info, data, reply, m
            str connect_host
            int connect_port

        # establish connection to appropriate host/port
        if use_proxy:
            if not use_tcps:
                errors._raise_err(errors.ERR_HTTPS_PROXY_REQUIRES_TCPS)
            connect_host = address.https_proxy
            connect_port = address.https_proxy_port
        else:
            connect_host = host
            connect_port = port
            if not use_tcps and (params._token is not None
                    or params.access_token_callback is not None):
                errors._raise_err(errors.ERR_ACCESS_TOKEN_REQUIRES_TCPS)
        transport, protocol = await self._read_buf._loop.create_connection(
            lambda: self,
            connect_host,
            connect_port
        )

        # complete connection through proxy, if applicable
        if use_proxy:
            data = f"CONNECT {host}:{port} HTTP/1.0\r\n\r\n"
            transport.write(data.encode())
            reply = transport.read(1024)
            m = re.search('HTTP/1.[01]\\s+(\\d+)\\s+', reply.decode())
            if m is None or m.groups()[0] != '200':
                errors._raise_err(errors.ERR_PROXY_FAILURE,
                                  response=reply.decode())

        # set socket on transport
        self._transport.set_from_socket(transport, params, description,
                                        address)

        # negotiate TLS, if applicable
        if use_tcps:
            self._transport.create_ssl_context(params, description, address)
            return await self._transport.negotiate_tls_async(self, description)

    async def _process_message(self, Message message):
        """
        Sends a message to the server and processes its response.
        """
        cdef:
            uint32_t timeout = message.conn_impl._call_timeout
            object timeout_obj = (timeout / 1000) or None
        try:
            coroutine = self._process_message_helper(message)
            await asyncio.wait_for(coroutine, timeout_obj)
        except asyncio.TimeoutError:
            try:
                coroutine = self._process_timeout_helper(message, timeout)
                await asyncio.wait_for(coroutine, timeout_obj)
            except asyncio.TimeoutError:
                self._force_close()
                errors._raise_err(errors.ERR_CONNECTION_CLOSED,
                                  "socket timed out while recovering from " \
                                  "previous socket timeout")
            raise
        except:
            if not self._in_connect \
                    and self._write_buf._packet_sent \
                    and self._read_buf._transport is not None:
                self._send_marker(self._write_buf, TNS_MARKER_TYPE_BREAK)
                await self._reset()
            raise
        if message.flush_out_binds:
            self._write_buf.start_request(TNS_PACKET_TYPE_DATA)
            self._write_buf.write_uint8(TNS_MSG_TYPE_FLUSH_OUT_BINDS)
            self._write_buf.end_request()
            await self._receive_packet(message)
            message.process(self._read_buf)
        if self._break_in_progress:
            try:
                coroutine = self._receive_packet(message)
                await asyncio.wait_for(coroutine, timeout_obj)
            except asyncio.TimeoutError:
                self._force_close()
                errors._raise_err(errors.ERR_CONNECTION_CLOSED,
                                  "socket timed out while awaiting break " \
                                  "response from server")
            message.process(self._read_buf)
            self._break_in_progress = False
        self._process_call_status(message.conn_impl, message.call_status)
        if message.error_occurred:
            if message.retry:
                message.error_occurred = False
                return await self._process_message(message)
            message._check_and_raise_exception()

    async def _process_message_helper(self, Message message):
        """
        Helper routine that is called to process a message within a timeout.
        """
        self._read_buf.reset_packets()
        message.send(self._write_buf)
        await self._receive_packet(message, check_request_boundary=True)
        while True:
            try:
                message.process(self._read_buf)
                break
            except OutOfPackets:
                await self._receive_packet(message)
                self._read_buf.restore_point()

    async def _process_single_message(self, Message message):
        """
        Process a single message within a request.
        """
        message.preprocess()
        async with self._request_lock:
            await self._process_message(message)
            if message.resend:
                await self._process_message(message)
        await message.postprocess_async()

    async def _process_timeout_helper(self, Message message, uint32_t timeout):
        """
        Helper routine that is called to process a timeout.
        """
        self._break_external()
        await self._receive_packet(message)
        self._break_in_progress = False
        errors._raise_err(errors.ERR_CALL_TIMEOUT_EXCEEDED, timeout=timeout)

    async def _receive_packet(self, Message message,
                              bint check_request_boundary=False,
                              bint in_pipeline=False):
        cdef:
            ReadBuffer buf = self._read_buf
            uint16_t refuse_message_len
            const char_type* ptr
        buf._check_request_boundary = \
                check_request_boundary and self._caps.supports_end_of_response
        await buf.wait_for_packets_async()
        buf._check_request_boundary = False
        if buf._current_packet.packet_type == TNS_PACKET_TYPE_MARKER:
            await self._reset(in_pipeline)
        elif buf._current_packet.packet_type == TNS_PACKET_TYPE_REFUSE:
            self._write_buf._packet_sent = False
            buf.skip_raw_bytes(2)
            buf.read_uint16(&refuse_message_len)
            if refuse_message_len == 0:
                message.error_info.message = None
            else:
                ptr = buf.read_raw_bytes(refuse_message_len)
                message.error_info.message = ptr[:refuse_message_len].decode()

    async def _reset(self, bint in_pipeline=False):
        cdef:
            uint8_t marker_type, packet_type
            uint8_t expected_marker_type

        # send reset marker, if not in a pipeline
        # NOTE: the expected marker type is different inside a pipeline!
        if in_pipeline:
            expected_marker_type = TNS_MARKER_TYPE_BREAK
        else:
            expected_marker_type = TNS_MARKER_TYPE_RESET
            self._send_marker(self._write_buf, TNS_MARKER_TYPE_RESET)

        # read and discard all packets until a reset marker is received
        while True:
            packet_type = self._read_buf._current_packet.packet_type
            if packet_type == TNS_PACKET_TYPE_MARKER:
                self._read_buf.skip_raw_bytes(2)
                self._read_buf.read_ub1(&marker_type)
                if marker_type == expected_marker_type:
                    break
            await self._read_buf.wait_for_packets_async()

        # read error packet; first skip as many marker packets as may be sent
        # by the server; if the server doesn't handle out-of-band breaks
        # properly, some quit immediately and others send multiple reset
        # markers (this addresses both situations without resulting in strange
        # errors)
        while packet_type == TNS_PACKET_TYPE_MARKER:
            await self._read_buf.wait_for_packets_async()
            packet_type = self._read_buf._current_packet.packet_type
        self._break_in_progress = False

    def connection_lost(self, exc):
        """
        Called when a connection has been lost. The presence of an exception
        indicates an abornmal loss of the connection. If in the process of
        establishing a connection, losing the connection is ignored since this
        can happen normally when a listener redirect is encountered.
        """
        if not self._in_connect:
            self._transport = None

    def data_received(self, data):
        """
        Called when data has been received on the transport.
        """
        cdef:
            bint notify_waiter = False
            Packet packet
        packet = self._transport.extract_packet(data)
        while packet is not None:
            self._read_buf._process_packet(packet, &notify_waiter, False)
            if notify_waiter:
                self._read_buf.notify_packet_received()
            packet = self._transport.extract_packet()

    async def end_pipeline(self, BaseThinConnImpl conn_impl, list messages,
                           bint continue_on_error):
        """
        Called when all messages for the pipeline have been sent to the
        database. An end pipeline message is sent to the database and then
        the responses to all of the messages are processed.
        """
        cdef Message message, end_message
        end_message = conn_impl._create_message(EndPipelineMessage)
        end_message.send(self._write_buf)
        try:
            for message in messages:
                try:
                    await self._receive_packet(message,
                                               check_request_boundary=True,
                                               in_pipeline=True)
                    message.preprocess()
                    message.process(self._read_buf)
                    self._process_call_status(conn_impl, message.call_status)
                    message._check_and_raise_exception()
                except Exception as e:
                    if not continue_on_error:
                        raise
                    message.pipeline_result_impl._capture_err(e)
            await self._receive_packet(end_message,
                                       check_request_boundary=True)
            end_message.process(self._read_buf)
            end_message._check_and_raise_exception()
        except:
            await self._reset()
            raise


class AsyncProtocol(BaseAsyncProtocol, asyncio.Protocol):
    pass



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3100 - Module for testing boolean variables
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_client_version() >= (12, 1), "unsupported client"
)
@unittest.skipUnless(
    test_env.get_server_version() >= (12, 1), "unsupported server"
)
class TestCase(test_env.BaseTestCase):
    def __test_bind_value_as_boolean(self, value):
        expected_result = str(bool(value)).upper()
        var = self.cursor.var(bool)
        var.setvalue(0, value)
        result = self.cursor.callfunc(
            "pkg_TestBooleans.GetStringRep", str, [var]
        )
        self.assertEqual(result, expected_result)

    def test_3100(self):
        "3100 - test binding in a False value"
        result = self.cursor.callfunc(
            "pkg_TestBooleans.GetStringRep", str, [False]
        )
        self.assertEqual(result, "FALSE")

    def test_3101(self):
        "3101 - test binding in a float as a boolean"
        self.__test_bind_value_as_boolean(0.0)
        self.__test_bind_value_as_boolean(1.0)

    def test_3102(self):
        "3102 - test binding in an integer as a boolean"
        self.__test_bind_value_as_boolean(0)
        self.__test_bind_value_as_boolean(1)

    def test_3103(self):
        "3103 - test binding in a null value"
        self.cursor.setinputsizes(None, bool)
        result = self.cursor.callfunc(
            "pkg_TestBooleans.GetStringRep", str, [None]
        )
        self.assertEqual(result, "NULL")

    def test_3104(self):
        "3104 - test binding out a boolean value (False)"
        result = self.cursor.callfunc(
            "pkg_TestBooleans.IsLessThan10", oracledb.DB_TYPE_BOOLEAN, [15]
        )
        self.assertFalse(result)

    def test_3105(self):
        "3105 - test binding out a boolean value (True)"
        result = self.cursor.callfunc(
            "pkg_TestBooleans.IsLessThan10", bool, [5]
        )
        self.assertTrue(result)

    def test_3106(self):
        "3106 - test binding in a string as a boolean"
        self.__test_bind_value_as_boolean("")
        self.__test_bind_value_as_boolean("0")

    def test_3107(self):
        "3107 - test binding in a True value"
        result = self.cursor.callfunc(
            "pkg_TestBooleans.GetStringRep", str, [True]
        )
        self.assertEqual(result, "TRUE")

    def test_3108(self):
        "3108 - test binding out a boolean value (None)"
        result = self.cursor.callfunc(
            "pkg_TestBooleans.TestOutValueNull", bool
        )
        self.assertIsNone(result)

    @unittest.skipUnless(
        test_env.get_client_version() >= (23, 1), "unsupported client"
    )
    @unittest.skipUnless(
        test_env.get_server_version() >= (23, 1), "unsupported server"
    )
    def test_3109(self):
        "3109 - test binding and fetching boolean with 23ai"
        for value in (True, False):
            with self.subTest(value=value):
                self.cursor.execute("select not :1 from dual", [value])
                (fetched_value,) = self.cursor.fetchone()
                self.assertIsInstance(fetched_value, bool)
                self.assertEqual(fetched_value, not value)

    @unittest.skipUnless(
        test_env.get_client_version() >= (23, 1), "unsupported client"
    )
    @unittest.skipUnless(
        test_env.get_server_version() >= (23, 1), "unsupported server"
    )
    def test_3110(self):
        "3110 - test binding and fetching string literals that represent True"
        self.cursor.execute("truncate table TestBooleans")
        true_values = ["true", "yes", "on", "1", "t", "y"]
        self.cursor.executemany(
            "insert into TestBooleans values (:1, :2, :3)",
            [(i, v, v) for i, v in enumerate(true_values)],
        )
        self.cursor.execute(
            "select BooleanCol1, BooleanCol2 from TestBooleans order by IntCol"
        )
        expected_values = [(True, True) for _ in true_values]
        self.assertEqual(self.cursor.fetchall(), expected_values)

    @unittest.skipUnless(
        test_env.get_client_version() >= (23, 1), "unsupported client"
    )
    @unittest.skipUnless(
        test_env.get_server_version() >= (23, 1), "unsupported server"
    )
    def test_3111(self):
        "3111 - test binding and fetching string literals that represent False"
        self.cursor.execute("truncate table TestBooleans")
        false_values = ["false", "no", "off", "0", "f", "n"]
        self.cursor.executemany(
            "insert into TestBooleans values (:1, :2, :3)",
            [(i, v, v) for i, v in enumerate(false_values)],
        )
        self.cursor.execute(
            "select BooleanCol1, BooleanCol2 from TestBooleans order by IntCol"
        )
        expected_value = [(False, False) for _ in range(len(false_values))]
        self.assertEqual(self.cursor.fetchall(), expected_value)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6600 - Module for testing oracledb.defaults
"""

import decimal
import tempfile

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def test_6600(self):
        "6600 - test setting defaults.arraysize"
        with test_env.DefaultsContextManager("arraysize", 50):
            conn = test_env.get_connection()
            cursor = conn.cursor()
            self.assertEqual(cursor.arraysize, oracledb.defaults.arraysize)

    def test_6601(self):
        "6601 - test getting decimals with defaults.fetch_decimals=True"
        with test_env.DefaultsContextManager("fetch_decimals", True):
            self.cursor.execute("select 9 from dual")
            (result,) = self.cursor.fetchone()
            self.assertIsInstance(result, decimal.Decimal)

    def test_6602(self):
        "6602 - test getting string lob with defaults.fetch_lobs=False"
        with test_env.DefaultsContextManager("fetch_lobs", False):
            self.cursor.execute("select to_clob('Hello world') from dual")
            (result,) = self.cursor.fetchone()
            self.assertIsInstance(result, str)

    def test_6603(self):
        "6603 - test setting defaults.prefetchrows"
        with test_env.DefaultsContextManager("prefetchrows", 20):
            conn = test_env.get_connection()
            cursor = conn.cursor()
            self.assertEqual(
                cursor.prefetchrows, oracledb.defaults.prefetchrows
            )

    def test_6604(self):
        "6604 - test setting defaults.stmtcachesize (pool)"
        with test_env.DefaultsContextManager("stmtcachesize", 40):
            pool = test_env.get_pool()
            self.assertEqual(
                pool.stmtcachesize, oracledb.defaults.stmtcachesize
            )
            conn = pool.acquire()
            self.assertEqual(
                conn.stmtcachesize, oracledb.defaults.stmtcachesize
            )

    def test_6605(self):
        "6605 - test setting defaults.stmtcachesize (standalone connection)"
        with test_env.DefaultsContextManager("stmtcachesize", 50):
            conn = test_env.get_connection()
            self.assertEqual(
                conn.stmtcachesize, oracledb.defaults.stmtcachesize
            )

    def test_6606(self):
        "6606 - fetch_lobs does not affect LOBS returned as OUT binds"
        with test_env.DefaultsContextManager("fetch_lobs", False):
            var = self.cursor.var(oracledb.DB_TYPE_CLOB)
            self.cursor.execute(
                "begin :value := to_clob('test clob'); end;",
                value=var,
            )
            self.assertIsInstance(var.getvalue(), oracledb.LOB)

    def test_6607(self):
        "6607 - test setting defaults.config_dir"
        with tempfile.TemporaryDirectory() as temp_dir:
            with test_env.DefaultsContextManager("config_dir", temp_dir):
                self.assertEqual(oracledb.defaults.config_dir, temp_dir)
                params = oracledb.ConnectParams()
                self.assertEqual(params.config_dir, temp_dir)

    def test_6608(self):
        "6608 - test setting defaults.stmtcachesize (ConnectParams)"
        with test_env.DefaultsContextManager("stmtcachesize", 50):
            params = oracledb.ConnectParams()
            self.assertEqual(
                params.stmtcachesize, oracledb.defaults.stmtcachesize
            )


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# queue.pyx
#
# Cython file defining the base Queue implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseQueueImpl:

    def deq_many(self, uint32_t max_num_messages):
        errors._raise_not_supported("dequeuing multiple messages")

    def deq_one(self):
        errors._raise_not_supported("dequeuing a single message")

    def enq_many(self, list props_impls):
        errors._raise_not_supported("enqueueing multiple messages")

    def enq_one(self, BaseMsgPropsImpl props_impl):
        errors._raise_not_supported("enqueuing a single message")

    def initialize(self, BaseConnImpl conn_impl, str name,
                   BaseDbObjectImpl payload_type, bint is_json):
        errors._raise_not_supported("initializing a queue")


cdef class BaseDeqOptionsImpl:

    def get_condition(self):
        errors._raise_not_supported("getting the condition")

    def get_consumer_name(self):
        errors._raise_not_supported("getting the consumer name")

    def get_correlation(self):
        errors._raise_not_supported("getting the correlation")

    def get_message_id(self):
        errors._raise_not_supported("getting the message id")

    def get_mode(self):
        errors._raise_not_supported("getting the mode")

    def get_navigation(self):
        errors._raise_not_supported("getting the navigation")

    def get_transformation(self):
        errors._raise_not_supported("getting the transformation")

    def get_visibility(self):
        errors._raise_not_supported("getting the visibility")

    def get_wait(self):
        errors._raise_not_supported("getting the wait time")

    def set_condition(self, str value):
        errors._raise_not_supported("setting the condition")

    def set_consumer_name(self, str value):
        errors._raise_not_supported("setting the consumer name")

    def set_correlation(self, str value):
        errors._raise_not_supported("setting the correlation")

    def set_delivery_mode(self, uint16_t value):
        errors._raise_not_supported("setting the delivery mode")

    def set_mode(self, uint32_t value):
        errors._raise_not_supported("setting the mode")

    def set_message_id(self, bytes value):
        errors._raise_not_supported("setting the message id")

    def set_navigation(self, uint32_t value):
        errors._raise_not_supported("setting the navigation")

    def set_transformation(self, str value):
        errors._raise_not_supported("setting the transformation")

    def set_visibility(self, uint32_t value):
        errors._raise_not_supported("setting the visibility")

    def set_wait(self, uint32_t value):
        errors._raise_not_supported("setting the wait time")


cdef class BaseEnqOptionsImpl:

    def get_transformation(self):
        errors._raise_not_supported("getting the transformation")

    def get_visibility(self):
        errors._raise_not_supported("getting the visibility")

    def set_delivery_mode(self, uint16_t value):
        errors._raise_not_supported("setting the delivery mode")

    def set_transformation(self, str value):
        errors._raise_not_supported("setting the transformation")

    def set_visibility(self, uint32_t value):
        errors._raise_not_supported("setting the visibility")


cdef class BaseMsgPropsImpl:

    def get_num_attempts(self):
        errors._raise_not_supported("getting the number of attempts")

    def get_correlation(self):
        errors._raise_not_supported("getting the correlation")

    def get_delay(self):
        errors._raise_not_supported("getting the delay")

    def get_delivery_mode(self):
        errors._raise_not_supported("getting the delivery mode")

    def get_enq_time(self):
        errors._raise_not_supported("getting the enqueue time")

    def get_exception_queue(self):
        errors._raise_not_supported("getting the name of the exception queue")

    def get_expiration(self):
        errors._raise_not_supported("getting the expiration")

    def get_message_id(self):
        errors._raise_not_supported("getting the message id")

    def get_priority(self):
        errors._raise_not_supported("getting the priority")

    def get_state(self):
        errors._raise_not_supported("getting the message state")

    def set_correlation(self, str value):
        errors._raise_not_supported("setting the correlation")

    def set_delay(self, int32_t value):
        errors._raise_not_supported("setting the delay")

    def set_exception_queue(self, str value):
        errors._raise_not_supported("setting the name of the exception queue")

    def set_expiration(self, int32_t value):
        errors._raise_not_supported("setting the expiration")

    def set_payload_bytes(self, bytes value):
        errors._raise_not_supported("setting the payload from bytes")

    def set_payload_object(self, BaseDbObjectImpl value):
        errors._raise_not_supported(
            "setting the payload from a database object"
        )

    def set_priority(self, int32_t value):
        errors._raise_not_supported("setting the priority")

    def set_recipients(self, list value):
        errors._raise_not_supported("setting recipients list")



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# enums.py
#
# Contains the enumerations of various constants used throughout the package.
# -----------------------------------------------------------------------------

import enum

from . import base_impl


class AuthMode(enum.IntFlag):
    DEFAULT = base_impl.AUTH_MODE_DEFAULT
    PRELIM = base_impl.AUTH_MODE_PRELIM
    SYSASM = base_impl.AUTH_MODE_SYSASM
    SYSBKP = base_impl.AUTH_MODE_SYSBKP
    SYSDBA = base_impl.AUTH_MODE_SYSDBA
    SYSDGD = base_impl.AUTH_MODE_SYSDGD
    SYSKMT = base_impl.AUTH_MODE_SYSKMT
    SYSOPER = base_impl.AUTH_MODE_SYSOPER
    SYSRAC = base_impl.AUTH_MODE_SYSRAC


class PipelineOpType(enum.IntFlag):
    CALL_FUNC = base_impl.PIPELINE_OP_TYPE_CALL_FUNC
    CALL_PROC = base_impl.PIPELINE_OP_TYPE_CALL_PROC
    COMMIT = base_impl.PIPELINE_OP_TYPE_COMMIT
    EXECUTE = base_impl.PIPELINE_OP_TYPE_EXECUTE
    EXECUTE_MANY = base_impl.PIPELINE_OP_TYPE_EXECUTE_MANY
    FETCH_ALL = base_impl.PIPELINE_OP_TYPE_FETCH_ALL
    FETCH_MANY = base_impl.PIPELINE_OP_TYPE_FETCH_MANY
    FETCH_ONE = base_impl.PIPELINE_OP_TYPE_FETCH_ONE


class PoolGetMode(enum.IntEnum):
    FORCEGET = base_impl.POOL_GETMODE_FORCEGET
    NOWAIT = base_impl.POOL_GETMODE_NOWAIT
    TIMEDWAIT = base_impl.POOL_GETMODE_TIMEDWAIT
    WAIT = base_impl.POOL_GETMODE_WAIT


class Purity(enum.IntEnum):
    DEFAULT = base_impl.PURITY_DEFAULT
    NEW = base_impl.PURITY_NEW
    SELF = base_impl.PURITY_SELF


class VectorFormat(enum.IntEnum):
    BINARY = base_impl.VECTOR_FORMAT_BINARY
    FLOAT32 = base_impl.VECTOR_FORMAT_FLOAT32
    FLOAT64 = base_impl.VECTOR_FORMAT_FLOAT64
    INT8 = base_impl.VECTOR_FORMAT_INT8


# provide aliases for all enumerated values
AUTH_MODE_DEFAULT = AuthMode.DEFAULT
AUTH_MODE_PRELIM = AuthMode.PRELIM
AUTH_MODE_SYSASM = AuthMode.SYSASM
AUTH_MODE_SYSBKP = AuthMode.SYSBKP
AUTH_MODE_SYSDBA = AuthMode.SYSDBA
AUTH_MODE_SYSDGD = AuthMode.SYSDGD
AUTH_MODE_SYSKMT = AuthMode.SYSKMT
AUTH_MODE_SYSOPER = AuthMode.SYSOPER
AUTH_MODE_SYSRAC = AuthMode.SYSRAC
PIPELINE_OP_TYPE_CALL_FUNC = PipelineOpType.CALL_FUNC
PIPELINE_OP_TYPE_CALL_PROC = PipelineOpType.CALL_PROC
PIPELINE_OP_TYPE_COMMIT = PipelineOpType.COMMIT
PIPELINE_OP_TYPE_EXECUTE = PipelineOpType.EXECUTE
PIPELINE_OP_TYPE_EXECUTE_MANY = PipelineOpType.EXECUTE_MANY
PIPELINE_OP_TYPE_FETCH_ALL = PipelineOpType.FETCH_ALL
PIPELINE_OP_TYPE_FETCH_MANY = PipelineOpType.FETCH_MANY
PIPELINE_OP_TYPE_FETCH_ONE = PipelineOpType.FETCH_ONE
POOL_GETMODE_FORCEGET = PoolGetMode.FORCEGET
POOL_GETMODE_NOWAIT = PoolGetMode.NOWAIT
POOL_GETMODE_TIMEDWAIT = PoolGetMode.TIMEDWAIT
POOL_GETMODE_WAIT = PoolGetMode.WAIT
PURITY_DEFAULT = Purity.DEFAULT
PURITY_NEW = Purity.NEW
PURITY_SELF = Purity.SELF
VECTOR_FORMAT_BINARY = VectorFormat.BINARY
VECTOR_FORMAT_FLOAT32 = VectorFormat.FLOAT32
VECTOR_FORMAT_FLOAT64 = VectorFormat.FLOAT64
VECTOR_FORMAT_INT8 = VectorFormat.INT8



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
7000 - Module for testing async connections shortcut methods
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_7000(self):
        "7000 - test execute() and fetchall()"
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.execute(
            "insert into TestTempTable (IntCol) values (:1)", [77]
        )
        await self.conn.execute(
            "insert into TestTempTable (IntCol) values (:val)", dict(val=15)
        )
        await self.conn.commit()

        res = await self.conn.fetchall(
            "select IntCol from TestTempTable order by IntCol"
        )
        self.assertEqual(res, [(15,), (77,)])

    async def test_7001(self):
        "7001 - test executemany()"
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.executemany(
            "insert into TestTempTable (IntCol) values (:1)", [(2,), (3,)]
        )
        await self.conn.executemany(
            "insert into TestTempTable (IntCol) values (:data)",
            [{"data": 4}, {"data": 5}],
        )
        await self.conn.commit()
        res = await self.conn.fetchall(
            "select IntCol from TestTempTable order by IntCol"
        )
        self.assertEqual(res, [(2,), (3,), (4,), (5,)])

    async def test_7002(self):
        "7002 - test fetchall() with arraysize"
        await self.conn.execute("truncate table TestTempTable")
        data = [(1,), (2,), (3,), (4,)]
        await self.conn.executemany(
            "insert into TestTempTable (IntCol) values (:value)",
            [{"value": i} for i, in data],
        )
        await self.conn.commit()

        await self.setup_round_trip_checker()
        res = await self.conn.fetchall(
            "select IntCol from TestTempTable order by IntCol", arraysize=1
        )
        self.assertEqual(res, data)
        await self.assertRoundTrips(5)

        res = await self.conn.fetchall(
            "select IntCol from TestTempTable order by IntCol",
            arraysize=len(data),
        )
        self.assertEqual(res, data)
        await self.assertRoundTrips(2)

    async def test_7003(self):
        "7003 - test fetchall() with rowfactory"
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test_7003')
            """
        )
        await self.conn.commit()

        column_names = ["INTCOL", "STRINGCOL1"]

        def rowfactory(*row):
            return dict(zip(column_names, row))

        res = await self.conn.fetchall(
            "select IntCol, StringCol1 from TestTempTable",
            rowfactory=rowfactory,
        )
        expected_value = [{"INTCOL": 1, "STRINGCOL1": "test_7003"}]
        self.assertEqual(res, expected_value)

    async def test_7004(self):
        "7004 - test fetchone()"
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.executemany(
            "insert into TestTempTable (IntCol) values (:1)", [(9,), (10,)]
        )
        await self.conn.commit()

        res = await self.conn.fetchone(
            "select IntCol from TestTempTable order by IntCol"
        )
        self.assertEqual(res, (9,))

        res = await self.conn.fetchone("select :1 from dual", [23])
        self.assertEqual(res, (23,))

        res = await self.conn.fetchone("select :val from dual", {"val": 5})
        self.assertEqual(res, (5,))

    async def test_7005(self):
        "7005 - test fetchone() with rowfactory"
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int, :str)
            """,
            [{"int": 3, "str": "Mac"}, {"int": 4, "str": "Doc"}],
        )
        await self.conn.commit()

        column_names = ["INT", "STRING"]

        def rowfactory(*row):
            return dict(zip(column_names, row))

        res = await self.conn.fetchone(
            "select IntCol, StringCol1 from TestTempTable order by IntCol",
            rowfactory=rowfactory,
        )
        self.assertEqual(res, {"INT": 3, "STRING": "Mac"})

    async def test_7006(self):
        "7006 - test fetchmany()"
        data = [(i,) for i in range(10)]
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.executemany(
            "insert into TestTempTable (IntCol) values (:1)", data
        )
        await self.conn.commit()
        res = await self.conn.fetchmany(
            "select IntCol from TestTempTable order by IntCol"
        )
        self.assertEqual(res, data)

        res = await self.conn.fetchmany("select :1 from dual", [1099])
        self.assertEqual(res, [(1099,)])

        res = await self.conn.fetchmany("select :val from dual", {"val": 366})
        self.assertEqual(res, [(366,)])

    async def test_7007(self):
        "7007 - test fetchmany() with num_rows"
        data = [(i,) for i in range(10)]
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.executemany(
            "insert into TestTempTable (IntCol) values (:1)", data
        )
        num_rows = 7
        res = await self.conn.fetchmany(
            "select IntCol from TestTempTable order by IntCol",
            num_rows=num_rows,
        )
        self.assertEqual(res, data[:num_rows])

    async def test_7008(self):
        "7008 - test fetchmany() with rowfactory and num_rows"
        conn = await test_env.get_connection_async()
        await conn.execute("truncate table TestTempTable")
        await conn.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int, :str)
            """,
            [{"int": 29, "str": "Feb"}, {"int": 4, "str": "Monday"}],
        )
        await conn.commit()

        column_names = ["INT", "STRING"]

        def rowfactory(*row):
            return dict(zip(column_names, row))

        res = await conn.fetchmany(
            "select IntCol, StringCol1 from TestTempTable order by IntCol",
            rowfactory=rowfactory,
        )
        expected_value = [
            {"INT": 4, "STRING": "Monday"},
            {"INT": 29, "STRING": "Feb"},
        ]
        self.assertEqual(res, expected_value)

        res = await conn.fetchmany(
            "select IntCol, StringCol1 from TestTempTable order by IntCol",
            rowfactory=rowfactory,
            num_rows=1,
        )
        self.assertEqual(res, [{"INT": 4, "STRING": "Monday"}])

    async def test_7009(self):
        "7009 - test callfunc()"
        # parameters
        res = await self.conn.callfunc(
            "func_Test", oracledb.NUMBER, ("Yes", 7)
        )
        self.assertEqual(res, 10)

        # keyword parameters
        kwargs = {"a_String": "Keyword", "a_ExtraAmount": 12}
        res = await self.conn.callfunc(
            "func_Test", oracledb.NUMBER, keyword_parameters=kwargs
        )
        self.assertEqual(res, 19)

        # paramters and keyword parameters
        kwargs = {"a_ExtraAmount": 25}
        res = await self.conn.callfunc(
            "func_Test", oracledb.NUMBER, ["Mixed"], kwargs
        )
        self.assertEqual(res, 30)

    async def test_7010(self):
        "7010 - test callproc() with parameters"
        var = self.cursor.var(oracledb.NUMBER)
        results = await self.conn.callproc("proc_Test", ("hi", 5, var))
        self.assertEqual(results, ["hi", 10, 2.0])

    async def test_7011(self):
        "7011 - test callproc() with keyword_parameters"
        in_out_value = self.cursor.var(oracledb.NUMBER)
        in_out_value.setvalue(0, 7)
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(
            a_InValue="Peace", a_InOutValue=in_out_value, a_OutValue=out_value
        )
        results = await self.conn.callproc("proc_Test", [], kwargs)
        self.assertEqual(results, [])
        self.assertEqual(in_out_value.getvalue(), 35)
        self.assertEqual(out_value.getvalue(), 5)

    async def test_7012(self):
        "7012 - test callproc() with parameters and keyword_parameters"
        in_out_value = self.cursor.var(oracledb.NUMBER)
        in_out_value.setvalue(0, 8)
        out_value = self.cursor.var(oracledb.NUMBER)
        kwargs = dict(a_InOutValue=in_out_value, a_OutValue=out_value)
        results = await self.conn.callproc("proc_Test", ["Input_7712"], kwargs)
        self.assertEqual(results, ["Input_7712"])
        self.assertEqual(in_out_value.getvalue(), 80)
        self.assertEqual(out_value.getvalue(), 10)

    async def test_7013(self):
        "7013 - test fetchmany() num_rows with 0 and negative values"
        data = [(i,) for i in range(10)]
        await self.conn.execute("truncate table TestTempTable")
        await self.conn.executemany(
            "insert into TestTempTable (IntCol) values (:1)", data
        )
        await self.conn.commit()
        for num_rows in (0, -1, -10):
            res = await self.conn.fetchmany(
                "select IntCol from TestTempTable",
                num_rows=num_rows,
            )
            self.assertEqual(res, [])

    async def test_7014(self):
        "7014 - test shortcut methods with transaction_in_progress"
        await self.conn.execute("truncate table TestTempTable")
        self.assertFalse(self.conn.transaction_in_progress)
        await self.conn.execute(
            "insert into TestTempTable (IntCol) values (5)",
        )
        self.assertTrue(self.conn.transaction_in_progress)
        await self.conn.commit()
        self.assertFalse(self.conn.transaction_in_progress)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
5300 - Module for testing connections with asyncio
"""

import asyncio
import random
import string
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    requires_connection = False

    async def __connect_and_drop(self):
        """
        Connect to the database, perform a query and drop the connection.
        """
        await asyncio.sleep(0.1)
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            await cursor.execute("select count(*) from TestNumbers")
            (count,) = await cursor.fetchone()
            self.assertEqual(count, 10)

    async def __verify_fetched_data(self, connection):
        expected_data = [f"String {i + 1}" for i in range(10)]
        sql = "select StringCol from TestStrings order by IntCol"
        for i in range(5):
            with connection.cursor() as cursor:
                await cursor.execute(sql)
                fetched_data = [s async for s, in cursor]
                self.assertEqual(fetched_data, expected_data)

    async def __verify_attributes(self, connection, attr_name, value, sql):
        setattr(connection, attr_name, value)
        cursor = connection.cursor()
        await cursor.execute(sql)
        (result,) = await cursor.fetchone()
        self.assertEqual(result, value, f"{attr_name} value mismatch")

    async def test_5300(self):
        "5300 - simple connection to database"
        async with test_env.get_connection_async() as conn:
            self.assertEqual(
                conn.username, test_env.get_main_user(), "user name differs"
            )
            self.assertEqual(
                conn.dsn, test_env.get_connect_string(), "dsn differs"
            )
            self.assertEqual(conn.thin, test_env.get_is_thin())

    async def test_5303(self):
        "5303 - test connection end-to-end tracing attributes"
        async with test_env.get_connection_async() as conn:
            if not await self.is_on_oracle_cloud(conn):
                sql = """select dbop_name from v$sql_monitor
                         where sid = sys_context('userenv', 'sid')
                         and status = 'EXECUTING'"""
                await self.__verify_attributes(
                    conn, "dbop", "oracledb_dbop", sql
                )
            sql = "select sys_context('userenv', 'action') from dual"
            await self.__verify_attributes(
                conn, "action", "oracledb_Action", sql
            )
            await self.__verify_attributes(conn, "action", None, sql)
            sql = "select sys_context('userenv', 'module') from dual"
            await self.__verify_attributes(
                conn, "module", "oracledb_Module", sql
            )
            await self.__verify_attributes(conn, "module", None, sql)
            sql = "select sys_context('userenv', 'client_info') from dual"
            await self.__verify_attributes(
                conn, "clientinfo", "oracledb_cinfo", sql
            )
            await self.__verify_attributes(conn, "clientinfo", None, sql)
            sql = (
                "select sys_context('userenv', 'client_identifier') from dual"
            )
            await self.__verify_attributes(
                conn, "client_identifier", "oracledb_cid", sql
            )
            await self.__verify_attributes(
                conn, "client_identifier", None, sql
            )

    async def test_5304(self):
        "5304 - test use of autocommit"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            other_conn = await test_env.get_connection_async()
            other_cursor = other_conn.cursor()
            await cursor.execute("truncate table TestTempTable")
            await cursor.execute(
                "insert into TestTempTable (IntCol) values (1)"
            )
            await other_cursor.execute("select IntCol from TestTempTable")
            self.assertEqual(await other_cursor.fetchall(), [])
            conn.autocommit = True
            await cursor.execute(
                "insert into TestTempTable (IntCol) values (2)"
            )
            await other_cursor.execute(
                "select IntCol from TestTempTable order by IntCol"
            )
            self.assertEqual(await other_cursor.fetchall(), [(1,), (2,)])

    async def test_5305(self):
        "5305 - connection to database with bad connect string"
        with self.assertRaisesFullCode(
            "DPY-4000", "DPY-4026", "DPY-4027", "ORA-12154"
        ):
            await oracledb.connect_async("not a valid connect string!!")
        with self.assertRaisesFullCode("DPY-4000", "DPY-4001"):
            await oracledb.connect(
                test_env.get_main_user() + "@" + test_env.get_connect_string()
            )

    async def test_5306(self):
        "5306 - connection to database with bad password"
        with self.assertRaisesFullCode("ORA-01017"):
            await test_env.get_connection_async(
                password=test_env.get_main_password() + "X",
            )

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_5307(self):
        "5307 - test changing password"
        async with test_env.get_connection_async() as conn:
            if await self.is_on_oracle_cloud(conn):
                self.skipTest(
                    "passwords on Oracle Cloud are strictly controlled"
                )
            sys_random = random.SystemRandom()
            new_password = "".join(
                sys_random.choice(string.ascii_letters) for i in range(20)
            )
            await conn.changepassword(
                test_env.get_main_password(), new_password
            )
            conn = await test_env.get_connection_async(password=new_password)
            await conn.changepassword(
                new_password, test_env.get_main_password()
            )

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_5308(self):
        "5308 - test changing password to an invalid value"
        async with test_env.get_connection_async() as conn:
            if await self.is_on_oracle_cloud(conn):
                self.skipTest(
                    "passwords on Oracle Cloud are strictly controlled"
                )
            new_password = "1" * 1500
            with self.assertRaisesFullCode("ORA-01017", "ORA-00988"):
                await conn.changepassword(
                    test_env.get_main_password(), new_password
                )
            with self.assertRaisesFullCode(
                "ORA-01017", "ORA-28008", "ORA-00988"
            ):
                await conn.changepassword(
                    "incorrect old password", new_password
                )

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_5309(self):
        "5309 - test connecting with password containing / and @ symbols"
        async with test_env.get_connection_async() as conn:
            if await self.is_on_oracle_cloud(conn):
                self.skipTest(
                    "passwords on Oracle Cloud are strictly controlled"
                )
            sys_random = random.SystemRandom()
            chars = list(
                sys_random.choice(string.ascii_letters) for i in range(20)
            )
            chars[4] = "/"
            chars[8] = "@"
            new_password = "".join(chars)
            await conn.changepassword(
                test_env.get_main_password(), new_password
            )
            try:
                async with test_env.get_connection_async(
                    password=new_password
                ):
                    pass
            finally:
                await conn.changepassword(
                    new_password, test_env.get_main_password()
                )

    async def test_5310(self):
        "5310 - confirm an exception is raised after closing a connection"
        async with await test_env.get_connection_async() as conn:
            await conn.close()
            with self.assertRaisesFullCode("DPY-1001"):
                await conn.rollback()

    async def test_5312(self):
        "5312 - connection version is a string"
        async with test_env.get_connection_async() as conn:
            self.assertIsInstance(conn.version, str)

    async def test_5313(self):
        "5313 - connection rolls back before close"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            await cursor.execute("truncate table TestTempTable")
            other_conn = await test_env.get_connection_async()
            other_cursor = other_conn.cursor()
            await other_cursor.execute(
                "insert into TestTempTable (IntCol) values (1)"
            )
            other_cursor.close()
            await other_conn.close()
            await cursor.execute("select count(*) from TestTempTable")
            (count,) = await cursor.fetchone()
            self.assertEqual(count, 0)

    async def test_5315(self):
        "5315 - multiple connections to database with multiple threads"
        coroutines = [self.__connect_and_drop() for i in range(20)]
        await asyncio.gather(*coroutines)

    async def test_5316(self):
        "5316 - test string format of connection"
        async with test_env.get_connection_async() as conn:
            expected_value = "<oracledb.AsyncConnection to %s@%s>" % (
                test_env.get_main_user(),
                test_env.get_connect_string(),
            )
            self.assertEqual(str(conn), expected_value)

    async def test_5317(self):
        "5317 - test context manager - close"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            await cursor.execute("truncate table TestTempTable")
            await cursor.execute(
                "insert into TestTempTable (IntCol) values (1)"
            )
            await conn.commit()
            await cursor.execute(
                "insert into TestTempTable (IntCol) values (2)"
            )
        with self.assertRaisesFullCode("DPY-1001"):
            await conn.ping()
        conn = await test_env.get_connection_async()
        cursor = conn.cursor()
        await cursor.execute("select count(*) from TestTempTable")
        (count,) = await cursor.fetchone()
        self.assertEqual(count, 1)

    async def test_5318(self):
        "5318 - test connection attribute values"
        async with test_env.get_connection_async() as conn:
            if test_env.get_client_version() >= (12, 1):
                self.assertEqual(conn.ltxid, b"")
            self.assertIsNone(conn.current_schema)
            conn.current_schema = "test_schema"
            self.assertEqual(conn.current_schema, "test_schema")
            self.assertIsNone(conn.edition)
            conn.external_name = "test_external"
            self.assertEqual(conn.external_name, "test_external")
            conn.internal_name = "test_internal"
            self.assertEqual(conn.internal_name, "test_internal")
            conn.stmtcachesize = 30
            self.assertEqual(conn.stmtcachesize, 30)
            self.assertRaises(TypeError, conn.stmtcachesize, 20.5)
            self.assertRaises(TypeError, conn.stmtcachesize, "value")
            self.assertIsNone(conn.warning)

    async def test_5319(self):
        "5319 - test closed connection attribute values"
        conn = await test_env.get_connection_async()
        await conn.close()
        attr_names = [
            "current_schema",
            "edition",
            "external_name",
            "internal_name",
            "stmtcachesize",
            "warning",
        ]
        if test_env.get_client_version() >= (12, 1):
            attr_names.append("ltxid")
        for name in attr_names:
            with self.assertRaisesFullCode("DPY-1001"):
                getattr(conn, name)

    async def test_5320(self):
        "5320 - test connection ping makes a round trip"
        self.conn = test_env.get_connection_async()
        async with self.conn:
            await self.setup_round_trip_checker()
            await self.conn.ping()
            await self.assertRoundTrips(1)

    async def test_5325(self):
        "5325 - single connection to database with multiple threads"
        async with test_env.get_connection_async() as conn:
            coroutines = [self.__verify_fetched_data(conn) for i in range(3)]
            await asyncio.gather(*coroutines)

    @unittest.skipIf(
        test_env.get_is_implicit_pooling(),
        "sessions can change with implicit pooling",
    )
    async def test_5326(self):
        "5326 - test connection cancel"
        async with test_env.get_connection_async() as conn:
            sleep_proc_name = test_env.get_sleep_proc_name()

            async def perform_cancel():
                await asyncio.sleep(0.1)
                conn.cancel()

            async def perform_work():
                with self.assertRaises(oracledb.OperationalError):
                    with conn.cursor() as cursor:
                        await cursor.callproc(sleep_proc_name, [2])

            await asyncio.gather(perform_work(), perform_cancel())

            with conn.cursor() as cursor:
                await cursor.execute("select user from dual")
                (user,) = await cursor.fetchone()
                self.assertEqual(user, test_env.get_main_user().upper())

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    async def test_5327(self):
        "5327 - test changing password during connect"
        async with test_env.get_connection_async() as conn:
            if await self.is_on_oracle_cloud(conn):
                self.skipTest(
                    "passwords on Oracle Cloud are strictly controlled"
                )
            sys_random = random.SystemRandom()
            new_password = "".join(
                sys_random.choice(string.ascii_letters) for i in range(20)
            )
            conn = await test_env.get_connection_async(
                newpassword=new_password
            )
            conn = await test_env.get_connection_async(password=new_password)
            await conn.changepassword(
                new_password, test_env.get_main_password()
            )

    async def test_5328(self):
        "5328 - test use of autocommit during reexecute"
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        data_to_insert = [(1, "Test String #1"), (2, "Test String #2")]
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            other_conn = await test_env.get_connection_async()
            other_cursor = other_conn.cursor()
            await cursor.execute("truncate table TestTempTable")
            await cursor.execute(sql, data_to_insert[0])
            await other_cursor.execute(
                "select IntCol, StringCol1 from TestTempTable"
            )
            self.assertEqual(await other_cursor.fetchall(), [])
            conn.autocommit = True
            await cursor.execute(sql, data_to_insert[1])
            await other_cursor.execute(
                "select IntCol, StringCol1 from TestTempTable"
            )
            self.assertEqual(await other_cursor.fetchall(), data_to_insert)

    async def test_5329(self):
        "5329 - test current_schema is set properly"
        async with test_env.get_connection_async() as conn:
            self.assertIsNone(conn.current_schema)

            user = test_env.get_main_user().upper()
            proxy_user = test_env.get_proxy_user().upper()
            cursor = conn.cursor()
            await cursor.execute(
                f"alter session set current_schema={proxy_user}"
            )
            self.assertEqual(conn.current_schema, proxy_user)

            conn.current_schema = user
            self.assertEqual(conn.current_schema, user)

            await cursor.execute(
                "select sys_context('userenv', 'current_schema') from dual"
            )
            (result,) = await cursor.fetchone()
            self.assertEqual(result, user)

    async def test_5330(self):
        "5330 - test dbms_output package"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            test_string = "Testing DBMS_OUTPUT package"
            await cursor.callproc("dbms_output.enable")
            await cursor.callproc("dbms_output.put_line", [test_string])
            string_var = cursor.var(str)
            number_var = cursor.var(int)
            await cursor.callproc(
                "dbms_output.get_line", (string_var, number_var)
            )
            self.assertEqual(string_var.getvalue(), test_string)

    async def test_5331(self):
        "5331 - test connection call_timeout"
        async with test_env.get_connection_async() as conn:
            conn.call_timeout = 500  # milliseconds
            self.assertEqual(conn.call_timeout, 500)
            with self.assertRaisesFullCode("DPY-4011", "DPY-4024"):
                with conn.cursor() as cursor:
                    await cursor.callproc(test_env.get_sleep_proc_name(), [2])

    async def test_5332(self):
        "5332 - test Connection repr()"

        class MyConnection(oracledb.AsyncConnection):
            pass

        conn = await test_env.get_connection_async(conn_class=MyConnection)
        qual_name = conn.__class__.__qualname__
        expected_value = (
            f"<{__name__}.{qual_name} to {conn.username}@{conn.dsn}>"
        )
        self.assertEqual(repr(conn), expected_value)

        await conn.close()
        expected_value = f"<{__name__}.{qual_name} disconnected>"
        self.assertEqual(repr(conn), expected_value)

    async def test_5333(self):
        "5333 - test getting write-only attributes"
        async with test_env.get_connection_async() as conn:
            with self.assertRaises(AttributeError):
                conn.action
            with self.assertRaises(AttributeError):
                conn.dbop
            with self.assertRaises(AttributeError):
                conn.clientinfo
            with self.assertRaises(AttributeError):
                conn.econtext_id
            with self.assertRaises(AttributeError):
                conn.module
            with self.assertRaises(AttributeError):
                conn.client_identifier

    async def test_5334(self):
        "5334 - test error for invalid type for params and pool"
        pool = test_env.get_pool_async()
        await pool.close()
        with self.assertRaisesFullCode("DPY-1002"):
            await test_env.get_connection_async(pool=pool)
        with self.assertRaises(TypeError):
            await test_env.get_connection_async(
                pool="This isn't an instance of a pool"
            )
        with self.assertRaisesFullCode("DPY-2025"):
            await oracledb.connect_async(params={"number": 7})

    async def test_5335(self):
        "5335 - test connection instance name"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            await cursor.execute(
                """
                select upper(sys_context('userenv', 'instance_name'))
                from dual
                """
            )
            (instance_name,) = await cursor.fetchone()
            self.assertEqual(conn.instance_name.upper(), instance_name)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    @unittest.skipIf(
        test_env.get_server_version() < (23, 0),
        "unsupported server",
    )
    async def test_5337(self):
        "5337 - test maximum allowed length for password"
        async with test_env.get_connection_async() as conn:
            if await self.is_on_oracle_cloud(conn):
                self.skipTest(
                    "passwords on Oracle Cloud are strictly controlled"
                )

            original_password = test_env.get_main_password()
            new_password_32 = "a" * 32
            await conn.changepassword(original_password, new_password_32)
            conn = await test_env.get_connection_async(
                password=new_password_32
            )

            new_password_1024 = "a" * 1024
            await conn.changepassword(new_password_32, new_password_1024)
            conn = await test_env.get_connection_async(
                password=new_password_1024
            )
            await conn.changepassword(new_password_1024, original_password)

            new_password_1025 = "a" * 1025
            with self.assertRaisesFullCode("ORA-28218", "ORA-00972"):
                await conn.changepassword(original_password, new_password_1025)

    async def test_5338(self):
        "5338 - test getting db_name"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            await cursor.execute("select name from V$DATABASE")
            (db_name,) = await cursor.fetchone()
            self.assertEqual(conn.db_name.upper(), db_name.upper())

    async def test_5339(self):
        "5339 - test getting max_open_cursors"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            await cursor.execute(
                "select value from V$PARAMETER where name='open_cursors'"
            )
            (max_open_cursors,) = await cursor.fetchone()
            self.assertEqual(conn.max_open_cursors, int(max_open_cursors))

    async def test_5340(self):
        "5340 - test getting service_name"
        async with test_env.get_connection_async() as conn:
            cursor = conn.cursor()
            await cursor.execute(
                "select sys_context('userenv', 'service_name') from dual"
            )
            (service_name,) = await cursor.fetchone()
            self.assertEqual(conn.service_name, service_name)

    async def test_5341(self):
        "5341 - test transaction_in_progress"
        async with test_env.get_connection_async() as conn:
            self.assertFalse(conn.transaction_in_progress)

            cursor = conn.cursor()
            await cursor.execute("truncate table TestTempTable")
            self.assertFalse(conn.transaction_in_progress)

            await cursor.execute(
                "insert into TestTempTable (IntCol) values (1)"
            )
            self.assertTrue(conn.transaction_in_progress)

            await conn.commit()
            self.assertFalse(conn.transaction_in_progress)

    async def test_5342(self):
        "5342 - test getting db_domain"
        async with test_env.get_connection_async() as conn:
            (db_domain,) = await conn.fetchone(
                "select value from V$PARAMETER where name='db_domain'"
            )
            self.assertEqual(conn.db_domain, db_domain)

    async def test_5343(self):
        "5343 - test connection with invalid conn_class"
        with self.assertRaisesFullCode("DPY-2023"):
            await test_env.get_connection_async(
                conn_class=oracledb.ConnectionPool
            )

    async def test_5344(self):
        "5344 - test connection with an invalid pool"
        with self.assertRaises(TypeError):
            await oracledb.connect_async(pool="not a pool object")


if __name__ == "__main__":
    test_env.run_test_cases()
